"""SSH service module."""

from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

from redux import AutorunOptions
from ubo_gui.constants import WARNING_COLOR
from ubo_gui.menu.types import ActionItem, HeadlessMenu, Item, Menu

from ubo_app.colors import RUNNING_COLOR, STOPPED_COLOR
from ubo_app.store.core.types import RegisterSettingAppAction, SettingsCategory
from ubo_app.store.main import store
from ubo_app.store.services.ssh import SSHClearEnabledStateAction, SSHUpdateStateAction
from ubo_app.utils.async_ import create_task
from ubo_app.utils.monitor_unit import is_unit_enabled, monitor_unit
from ubo_app.utils.server import send_command

if TYPE_CHECKING:
    from collections.abc import Sequence

    from ubo_app.store.services.ssh import SSHState


def start_ssh_service() -> None:
    """Start the SSH service."""
    create_task(send_command('service', 'ssh', 'start'))


def stop_ssh_service() -> None:
    """Stop the SSH service."""
    create_task(send_command('service', 'ssh', 'stop'))


def enable_ssh_service() -> None:
    """Enable the SSH service."""

    async def act() -> None:
        store.dispatch(SSHClearEnabledStateAction())
        await send_command('service', 'ssh', 'enable')
        await asyncio.sleep(5)
        await check_is_ssh_enabled()

    create_task(act())


def disable_ssh_service() -> None:
    """Disable the SSH service."""

    async def act() -> None:
        store.dispatch(SSHClearEnabledStateAction())
        await send_command('service', 'ssh', 'disable')
        await asyncio.sleep(5)
        await check_is_ssh_enabled()

    create_task(act())


@store.autorun(lambda state: state.ssh)
def ssh_items(state: SSHState) -> Sequence[Item]:
    """Get the SSH menu items."""
    return [
        ActionItem(
            label='Stop' if state.is_active else 'Start',
            icon='󰓛' if state.is_active else '󰐊',
            action=stop_ssh_service if state.is_active else start_ssh_service,
        ),
        Item(
            label='...',
            icon='',
        )
        if state.is_enabled is None
        else ActionItem(
            label='Disable',
            icon=f'[color={RUNNING_COLOR}]󰯄[/color]',
            action=disable_ssh_service,
        )
        if state.is_enabled
        else ActionItem(
            label='Enable',
            icon=f'[color={STOPPED_COLOR}]󰯅[/color]',
            action=enable_ssh_service,
        ),
    ]


@store.autorun(
    lambda state: state.ssh,
    options=AutorunOptions(default_value=f'[color={WARNING_COLOR}][/color]'),
)
def ssh_icon(state: SSHState) -> str:
    """Get the SSH icon."""
    return (
        f'[color={RUNNING_COLOR}]󰪥[/color]'
        if state.is_active
        else f'[color={STOPPED_COLOR}]󰝦[/color]'
    )


@store.autorun(lambda state: state.ssh)
def ssh_title(_: SSHState) -> str:
    """Get the SSH title."""
    return ssh_icon() + ' SSH'


async def check_is_ssh_enabled() -> None:
    """Check if the SSH service is enabled."""
    if await is_unit_enabled('ssh'):
        store.dispatch(SSHUpdateStateAction(is_enabled=True))
    else:
        store.dispatch(SSHUpdateStateAction(is_enabled=False))


def open_ssh_menu() -> Menu:
    """Open the SSH menu."""
    create_task(check_is_ssh_enabled())

    return HeadlessMenu(title=ssh_title, items=ssh_items)


def init_service() -> None:
    """Initialize the SSH service."""
    store.dispatch(
        RegisterSettingAppAction(
            priority=1,
            category=SettingsCategory.REMOTE,
            menu_item=ActionItem(
                label='SSH',
                icon=ssh_icon,
                action=open_ssh_menu,
            ),
        ),
    )

    create_task(check_is_ssh_enabled())
    create_task(
        monitor_unit(
            'ssh.service',
            lambda status: store.dispatch(
                SSHUpdateStateAction(
                    is_active=status in ('active', 'activating', 'reloading'),
                ),
            ),
        ),
    )



"""Settings state module."""

from __future__ import annotations

from dataclasses import field
from enum import StrEnum
from pathlib import Path

from immutable import Immutable
from redux import BaseEvent

from ubo_app.constants import DEBUG_BETA_VERSIONS, DEBUG_PDB_SIGNAL, DEBUG_VISUAL
from ubo_app.store.core.types import MainAction
from ubo_app.utils.persistent_store import read_from_persistent_store

ROOT_PATH = Path(__file__).parent.parent.parent


class ServicesStatus(StrEnum):
    """Services status enum."""

    LOADING = 'loading'
    READY = 'ready'


class SettingsAction(MainAction):
    """Settings action."""


class SettingsTogglePdbSignalAction(SettingsAction):
    """Toggle PDB signal action."""


class SettingsToggleVisualDebugAction(SettingsAction):
    """Toggle visual debug mode action."""


class SettingsToggleBetaVersionsAction(SettingsAction):
    """Toggle beta versions action."""


class SettingsSetServicesAction(SettingsAction):
    """Start service action."""

    services: dict[str, ServiceState]
    gap_duration: float


class SettingsServiceAction(SettingsAction):
    """Service action."""

    service_id: str


class SettingsStartServiceAction(SettingsServiceAction):
    """Start service action."""


class SettingsStopServiceAction(SettingsServiceAction):
    """Stop service action."""


class SettingsServiceSetStatusAction(SettingsServiceAction):
    """Set service status action."""

    is_active: bool


class SettingsServiceSetIsEnabledAction(SettingsServiceAction):
    """Set service enabled action."""

    is_enabled: bool


class SettingsServiceSetLogLevelAction(SettingsServiceAction):
    """Set service log level action."""

    log_level: int


class SettingsServiceSetShouldRestartAction(SettingsServiceAction):
    """Set service should restart action."""

    should_auto_restart: bool


class SettingsReportServiceErrorAction(SettingsServiceAction):
    """Report service error action."""

    error: ErrorReport


class SettingsClearServiceErrorsAction(SettingsServiceAction):
    """Clear service errors action."""


class SettingsEvent(BaseEvent):
    """Settings event."""


class SettingsServiceEvent(SettingsEvent):
    """Service event."""

    service_id: str


class SettingsStartServiceEvent(SettingsServiceEvent):
    """Start service event."""

    delay: float = 0


class SettingsStopServiceEvent(SettingsServiceEvent):
    """Stop service event."""


class ErrorReport(Immutable):
    """Error report."""

    timestamp: float
    message: str


class ServiceState(Immutable):
    """Service state."""

    id: str
    label: str
    is_active: bool
    is_enabled: bool
    log_level: int
    should_auto_restart: bool
    errors: list[ErrorReport] = field(default_factory=list)


class SettingsState(Immutable):
    """Settings state."""

    pdb_signal: bool = field(
        default_factory=lambda: read_from_persistent_store(
            'settings:pdb_signal',
            default=DEBUG_PDB_SIGNAL,
        ),
    )
    visual_debug: bool = field(
        default_factory=lambda: read_from_persistent_store(
            'settings:visual_debug',
            default=DEBUG_VISUAL,
        ),
    )
    beta_versions: bool = field(
        default_factory=lambda: read_from_persistent_store(
            'settings:beta_versions',
            default=DEBUG_BETA_VERSIONS,
        ),
    )
    services: dict[str, ServiceState] = field(default_factory=dict)
    services_status: ServicesStatus = ServicesStatus.LOADING



"""File system utilities for handling file paths and sizes."""

from __future__ import annotations


def human_readable_size(size: float, decimal_places: int = 2) -> str:
    """Convert a size in bytes to a human-readable format."""
    for unit in ['bytes', 'KB', 'MB', 'GB', 'TB']:
        if size < 2**10:
            return (
                f'{size} {unit}'
                if unit == 'bytes'
                else f'{size:.{decimal_places}f} {unit}'
            )
        size /= 2.0**10

    return f'{size:.{decimal_places}f} PB'



# ruff: noqa: D100, D103
from __future__ import annotations

from dataclasses import replace

from redux import (
    CompleteReducerResult,
    InitAction,
    InitializationActionError,
    ReducerResult,
)

from ubo_app.store.services.users import (
    UsersAction,
    UsersCreateUserAction,
    UsersCreateUserEvent,
    UsersDeleteUserAction,
    UsersDeleteUserEvent,
    UsersEvent,
    UsersResetPasswordAction,
    UsersResetPasswordEvent,
    UsersSetUsersAction,
    UsersState,
)


def reducer(
    state: UsersState | None,
    action: UsersAction | InitAction,
) -> ReducerResult[UsersState, UsersAction, UsersEvent]:
    if state is None:
        if isinstance(action, InitAction):
            return UsersState()
        raise InitializationActionError(action)

    match action:
        case UsersSetUsersAction():
            return replace(state, users=action.users)

        case UsersCreateUserAction():
            return CompleteReducerResult(
                state=state,
                events=[UsersCreateUserEvent()],
            )

        case UsersDeleteUserAction():
            return CompleteReducerResult(
                state=state,
                events=[UsersDeleteUserEvent(id=action.id)],
            )

        case UsersResetPasswordAction():
            return CompleteReducerResult(
                state=state,
                events=[UsersResetPasswordEvent(id=action.id)],
            )

        case _:
            return state



"""Ubo menu application."""

from __future__ import annotations

from typing import TYPE_CHECKING

from kivy.clock import mainthread
from kivy.graphics.context_instructions import Color
from kivy.graphics.vertex_instructions import Rectangle
from kivy.uix.widget import Widget
from redux import AutorunOptions
from typing_extensions import override
from ubo_gui.app import UboApp

from ubo_app.menu_app.menu_central import MenuAppCentral
from ubo_app.menu_app.menu_footer import MenuAppFooter
from ubo_app.menu_app.menu_header import MenuAppHeader
from ubo_app.store.main import store
from ubo_app.store.services.display import DisplayRedrawEvent

if TYPE_CHECKING:
    from ubo_app.utils.types import Subscriptions


class BlankOverlay(Widget):
    """Full-screen black overlay widget for blanked display."""

    def __init__(self: BlankOverlay, **kwargs: object) -> None:
        """Initialize the blank overlay."""
        super().__init__(**kwargs)
        with self.canvas:
            Color(0, 0, 0, 1)  # Black color
            self.rect = Rectangle(size=self.size, pos=self.pos)
        self.bind(size=self._update_rect, pos=self._update_rect)

    def _update_rect(self: BlankOverlay, *_args: object) -> None:
        """Update rectangle size and position."""
        self.rect.size = self.size
        self.rect.pos = self.pos


class MenuApp(MenuAppCentral, MenuAppFooter, MenuAppHeader, UboApp):
    """Menu application."""

    def __init__(self: MenuApp, **kwargs: object) -> None:
        """Initialize the application."""
        super().__init__(**kwargs)
        self.subscriptions: Subscriptions = []
        self.is_stopped = False
        self.blank_overlay: BlankOverlay | None = None
        self.saved_children: list[Widget] = []

    def set_visual_debug_mode(self: MenuApp, visual_debug: bool) -> None:  # noqa: FBT001
        """Set the visual debug mode."""
        self.root.show_update_regions = visual_debug

    @mainthread
    def handle_blank_state(self: MenuApp, is_blanked: bool) -> None:  # noqa: FBT001
        """Show or hide blank overlay based on blanked state."""
        if is_blanked:
            # Save current children and replace with blank overlay
            self.saved_children = list(self.root.children)
            self.root.clear_widgets()
            if self.blank_overlay is None:
                self.blank_overlay = BlankOverlay(size=self.root.size)
            self.root.add_widget(self.blank_overlay)
        elif self.saved_children:
            # Restore saved children
            self.root.clear_widgets()
            for child in reversed(self.saved_children):
                self.root.add_widget(child)
            self.saved_children = []

    def rerender(self: MenuApp) -> None:
        """Re-render the application."""
        self.root.previous_frame = None
        mainthread(self.root.process_frame)()

    @override
    def on_start(self: MenuApp) -> None:
        """Start the application."""
        from ubo_app.side_effects import setup_side_effects

        self.subscriptions = setup_side_effects()

        store.autorun(
            lambda state: state.settings.visual_debug,
            options=AutorunOptions(keep_ref=False),
        )(self.set_visual_debug_mode)

        # Monitor is_blanked state and show/hide overlay
        store.autorun(
            lambda state: (
                state.display.is_blanked if hasattr(state, 'display') else False
            ),
            options=AutorunOptions(keep_ref=False),
        )(self.handle_blank_state)

        store.subscribe_event(DisplayRedrawEvent, self.rerender, keep_ref=False)

    @override
    def stop(self, *largs: object) -> None:
        """Stop the application."""
        super().stop(*largs)
        self.is_stopped = True
        for cleanup in self.subscriptions:
            cleanup()



"""Test for AsyncEvictingQueue."""

from __future__ import annotations

import asyncio

from ubo_app.utils.async_evicting_queue import AsyncEvictingQueue


async def test_async_evicting_queue() -> None:
    """Run tests for the AsyncEvictingQueue."""
    q = AsyncEvictingQueue(3)

    # Test 1: Basic put and get
    await q.put(1)
    assert await q.get() == 1, 'Basic put/get failed'
    assert q.empty(), 'Queue should be empty'

    # Test 2: Async wait on empty queue
    async def delayed_put() -> None:
        await asyncio.sleep(0.1)
        await q.put(2)

    task = asyncio.create_task(delayed_put())
    assert await q.get() == 2, 'Async wait on get failed'
    await task

    # Test 3: Overflow discards oldest
    await q.put(1)
    await q.put(2)
    await q.put(3)
    await q.put(4)  # Should discard 1
    assert q.qsize() == 3, 'Queue size should be 3'
    assert await q.get() == 2, 'Oldest item not discarded'

    # Test 4: Full and empty checks
    await q.put(5)
    await q.put(6)
    assert q.full(), 'Queue should be full'
    await q.get()
    await q.get()
    await q.get()
    assert q.empty(), 'Queue should be empty'



# ruff: noqa: D100, D101
from __future__ import annotations

from immutable import Immutable
from redux import BaseAction, BaseEvent


class RPiConnectAction(BaseAction): ...


class RPiConnectEvent(BaseEvent): ...


class RPiConnectStartDownloadingAction(RPiConnectAction): ...


class RPiConnectDoneDownloadingAction(RPiConnectAction): ...


class RPiConnectSetPendingAction(RPiConnectAction): ...


class RPiConnectStatus(Immutable):
    screen_sharing_sessions: int | None
    remote_shell_sessions: int | None


class RPiConnectSetStatusAction(RPiConnectAction):
    is_installed: bool
    is_signed_in: bool | None
    status: RPiConnectStatus | None


class RPiConnectLoginEvent(RPiConnectEvent): ...


class RPiConnectUpdateServiceStateAction(RPiConnectAction):
    is_active: bool | None = None


class RPiConnectState(Immutable):
    is_downloading: bool = False
    is_active: bool = False
    is_installed: bool | None = None
    is_signed_in: bool | None = None
    status: RPiConnectStatus | None = None



# ruff: noqa: D100, D101, D102, D107
from __future__ import annotations

import asyncio
import pathlib
from typing import TYPE_CHECKING, cast

from kivy.lang.builder import Builder
from kivy.properties import BooleanProperty
from str_to_bool import str_to_bool
from wifi_manager import add_wireless_connection

from ubo_app.colors import SUCCESS_COLOR, WARNING_COLOR
from ubo_app.logger import logger
from ubo_app.store.core.types import CloseApplicationAction
from ubo_app.store.input.types import (
    AvailableInputDescription,
    InputFieldDescription,
    InputFieldType,
    InputMethod,
    QRCodeInputDescription,
    WebUIInputDescription,
)
from ubo_app.store.main import store
from ubo_app.store.services.notifications import (
    Chime,
    Notification,
    NotificationDisplayType,
    NotificationsAddAction,
    NotificationsClearByIdAction,
)
from ubo_app.store.services.speech_synthesis import ReadableInformation
from ubo_app.store.services.wifi import WiFiType, WiFiUpdateRequestAction
from ubo_app.utils.async_ import create_task
from ubo_app.utils.gui import UboPageWidget
from ubo_app.utils.input import ubo_input

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence

# Regular expression pattern
# WIFI:S:<SSID>;T:<WEP|WPA|blank>;P:<PASSWORD>;H:<true|false|blank>;;
BARCODE_PATTERN = (
    r'^WIFI:S:(?P<SSID>[^;]*);(?:T:(?P<Type>(?i:WEP|WPA|WPA2|nopass));)'
    r'?(?:P:(?P<Password>[^;]*);)?(?:H:(?P<Hidden>(?i:true|false));)?;?$|'
    r'^WIFI:T:(?P<Type_>(?i:WEP|WPA|WPA2|nopass));S:(?P<SSID_>[^;]*);'
    r'(?:P:(?P<Password_>[^;]*);)?(?:H:(?P<Hidden_>(?i:true|false|));)?;?$'
)
HOTSPOT_GRACE_TIME = 5
INPUT_DESCRIPTIONS: Sequence[AvailableInputDescription] = [
    QRCodeInputDescription(
        pattern=BARCODE_PATTERN,
        instructions=ReadableInformation(
            text='Go to your phone settings, choose QR code and hold it in '
            'front of the camera to scan it.',
            picovoice_text='Go to your phone settings, choose {QR|K Y UW AA R} '
            'code and hold it in front of the camera to scan it.',
        ),
    ),
    WebUIInputDescription(
        fields=[
            InputFieldDescription(
                name='SSID',
                label='SSID',
                type=InputFieldType.TEXT,
                description='The name of the WiFi network',
                required=True,
            ),
            InputFieldDescription(
                name='Password',
                label='Password',
                type=InputFieldType.PASSWORD,
                description='The password of the WiFi network',
                required=False,
            ),
            InputFieldDescription(
                name='Type',
                label='Type',
                type=InputFieldType.SELECT,
                description='The type of the WiFi network',
                default_value='WPA2',
                options=['WEP', 'WPA', 'WPA2', 'nopass'],
                required=False,
            ),
            InputFieldDescription(
                name='Hidden',
                label='Hidden',
                type=InputFieldType.CHECKBOX,
                description='Is the WiFi network hidden?',
                default_value='false',
                required=False,
            ),
        ],
    ),
]


async def input_wifi_connection(
    *,
    input_methods: tuple[InputMethod, ...] = (),
    on_creating: Callable[[], None] | None = None,
) -> None:
    """Input WiFi connection."""
    logger.debug('wifi connection input - start')
    try:
        _, result = await ubo_input(
            prompt='Enter WiFi connection',
            descriptions=[
                description
                for description in INPUT_DESCRIPTIONS
                if not input_methods or description.input_method in input_methods
            ],
        )
    except asyncio.CancelledError:
        logger.debug('wifi connection input - cancelled')
        return
    except Exception:
        logger.exception('wifi connection input - error')
        raise

    logger.debug('wifi connection input - result', extra={'result': result})

    if not result:
        logger.debug('wifi connection input - no result')
        return
    ssid = result.data.get('SSID') or result.data.get('SSID_')
    if ssid is None:
        logger.debug('wifi connection input - no ssid')
        return

    password = result.data.get('Password') or result.data.get('Password_')
    type = result.data.get('Type') or result.data.get('Type_')
    if type:
        type = type.upper()
    type = cast('WiFiType', type)
    hidden = str_to_bool(
        result.data.get('Hidden') or result.data.get('Hidden_') or 'false',
    )

    if not password:
        logger.debug('wifi connection input - no password')
        logger.warning('Password is required')
        return

    if on_creating:
        on_creating()

    if result.method is InputMethod.WEB_DASHBOARD:
        logger.debug(
            'wifi connection input - waiting for hotspot to go down',
            extra={'grace time': HOTSPOT_GRACE_TIME},
        )
        # Wait for hotspot to go down
        notification = Notification(
            id='wifi-wait-hotspot',
            title='Please wait!',
            content='To avoid interference we need to wait for the hotspot to go down.',
            display_type=NotificationDisplayType.STICKY,
            color=WARNING_COLOR,
            icon='󱋆',
        )
        store.dispatch(NotificationsAddAction(notification=notification))
        await asyncio.sleep(HOTSPOT_GRACE_TIME)
        store.dispatch(NotificationsClearByIdAction(id='wifi-wait-hotspot'))
        logger.debug(
            'wifi connection input - done waiting for hotspot to go down',
            extra={'grace time': HOTSPOT_GRACE_TIME},
        )

    logger.debug('wifi connection input - creating connection')
    try:
        await add_wireless_connection(
            ssid=ssid,
            password=password,
            type=type or WiFiType.NOPASS,
            hidden=hidden,
        )
    except Exception:
        logger.exception('wifi connection input - error while creating connection')
        raise

    logger.info(
        'Wireless connection created',
        extra={
            'ssid': ssid,
            'password': '<HIDDEN>' if password else '<NOT PROVIDED>',
            'type': type,
            'hidden': hidden,
        },
    )

    store.dispatch(
        WiFiUpdateRequestAction(reset=True),
        NotificationsAddAction(
            notification=Notification(
                title=f'"{ssid}" Added',
                content=f"""WiFi connection with ssid "{
                    ssid
                }" was added successfully""",
                display_type=NotificationDisplayType.FLASH,
                color=SUCCESS_COLOR,
                icon='󱛃',
                chime=Chime.ADD,
            ),
        ),
    )


class CreateWirelessConnectionPage(UboPageWidget):
    creating = BooleanProperty(defaultvalue=False)

    def __init__(
        self,
        input_methods: tuple[InputMethod, ...] = (),
        **kwargs: object,
    ) -> None:
        super().__init__(**kwargs)
        self.input_methods = input_methods
        create_task(self.create_wireless_connection())

    async def create_wireless_connection(self) -> None:
        await input_wifi_connection(
            on_creating=lambda: setattr(self, 'creating', True),
            input_methods=self.input_methods,
        )
        store.dispatch(CloseApplicationAction(application_instance_id=self.id))


Builder.load_file(
    pathlib.Path(__file__)
    .parent.joinpath('create_wireless_connection.kv')
    .resolve()
    .as_posix(),
)



"""Utilities for logging the output of failed processes."""

from asyncio.subprocess import Process

from redux import DispatchParameters

from ubo_app.store.services.notifications import (
    Importance,
    Notification,
    NotificationsAddAction,
)
from ubo_app.store.services.speech_synthesis import ReadableInformation
from ubo_app.utils.error_handlers import report_service_error


async def log_async_process(
    process: Process,
    *,
    title: str = 'Error',
    message: str,
) -> DispatchParameters:
    """Log the output of the process if it fails.

    Arguments:
    ---------
    process: Process - The process to log the output of.
    title: str - The title of the notification.
    message: str - The message to display in the notification.

    """
    if process.returncode != 0:
        logs = '\n'
        if process.stdout:
            logs += '------\nstdout:\n' + (await process.stdout.read()).decode()
        if process.stderr:
            logs += '------\nstderr:\n' + (await process.stderr.read()).decode()
        report_service_error(context={'message': logs})
        return (
            NotificationsAddAction(
                notification=Notification(
                    title=title,
                    content=message,
                    extra_information=ReadableInformation(
                        text=logs,
                        piper_text='',
                        picovoice_text='',
                    ),
                    importance=Importance.HIGH,
                ),
            ),
        )

    return []



# ruff: noqa: D100, D101, D102, D103, D105, D107
from __future__ import annotations

import asyncio
import contextlib
import ctypes
import importlib
import importlib.abc
import importlib.util
import inspect
import logging
import os
import signal
import sys
import threading
import traceback
import weakref
from collections import OrderedDict
from dataclasses import replace
from importlib.machinery import PathFinder, SourceFileLoader
from pathlib import Path
from typing import TYPE_CHECKING, ClassVar, cast

from redux import (
    CombineReducerRegisterAction,
    CombineReducerUnregisterAction,
    ReducerType,
)

from ubo_app.constants import (
    DATA_PATH,
    DEBUG_TASKS,
    DISABLED_SERVICES,
    ENABLED_SERVICES,
    GRPC_LISTEN_ADDRESS,
    GRPC_LISTEN_PORT,
    PACKAGE_NAME,
    SERVICES_LOOP_GRACE_PERIOD,
    SERVICES_PATH,
)
from ubo_app.logger import ThreadLevelFilter, logger
from ubo_app.store.settings.types import (
    ServiceState,
    SettingsServiceSetStatusAction,
    SettingsSetServicesAction,
    SettingsStartServiceEvent,
    SettingsStopServiceEvent,
)
from ubo_app.utils.error_handlers import STACKS, loop_exception_handler
from ubo_app.utils.service import ServiceUnavailableError, get_service

if TYPE_CHECKING:
    from collections.abc import Callable, Coroutine, Sequence
    from importlib.machinery import ModuleSpec
    from types import ModuleType

    from redux.basic_types import TaskCreatorCallback
    from ubo_handle import (  # pyright: ignore [reportMissingModuleSource]
        ReducerRegistrar,
        SetupFunction,
        SetupFunctionReturnType,
    )

    from ubo_app.utils.types import Subscriptions

SERVICES_BY_PATH: dict[Path, UboServiceThread] = {}
SERVICE_PATHS_BY_ID: dict[str, Path] = OrderedDict()
ROOT_PATH = Path(__file__).parent


class DisabledServiceError(Exception):
    """Raised when a service is disabled."""


# Customized module finder and module loader for ubo services to avoid mistakenly
# loading conflicting names in different services.
class UboModuleLoader(SourceFileLoader):
    cache: ClassVar[weakref.WeakValueDictionary[str, ModuleType]] = (
        weakref.WeakValueDictionary()
    )

    @property
    def _cache_id(self: UboModuleLoader) -> str:
        return f'{self.path}:{self.name}'

    def get_filename(self, name: str | None = None) -> str:
        return super().get_filename(name.split(':')[-1] if name else name)

    def create_module(self, spec: ModuleSpec) -> ModuleType | None:
        if self._cache_id in self.cache:
            return self.cache[self._cache_id]
        return super().create_module(spec)

    def exec_module(self, module: ModuleType) -> None:
        if self._cache_id in self.cache:
            return
        self.cache[self._cache_id] = module
        sys.modules[module.__name__] = module
        super().exec_module(module)


class UboServiceFinder(importlib.abc.MetaPathFinder):
    def find_spec(
        self,
        fullname: str,
        _: Sequence[str] | None,
        target: ModuleType | None = None,
    ) -> ModuleSpec | None:
        if fullname.startswith(PACKAGE_NAME):
            return None

        if fullname.startswith('/') and fullname in sys.modules:
            return sys.modules[fullname].__spec__

        try:
            service = get_service()
        except ServiceUnavailableError:
            pass
        else:
            if not service.is_alive():
                if service.is_started:
                    msg = 'No import is allowed after the service is finished.'
                    raise ImportError(msg)
                msg = (
                    'No import is allowed in the global scope of `ubo_handle.py` files.'
                )
                raise ImportError(msg)

            module_name = f'{service.service_uid}:{fullname}'

            if module_name in sys.modules:
                module = sys.modules[module_name]
                return module.__spec__

            fullname_parts = fullname.split('.')
            search_path = service.path.joinpath(*fullname_parts[:-1])
            spec = PathFinder.find_spec(
                fullname_parts[-1],
                [search_path.as_posix()],
                target,
            )
            if spec and spec.origin:
                spec.name = module_name
                spec.loader = UboModuleLoader(fullname, spec.origin)
            return spec

        return None


sys.meta_path.insert(0, UboServiceFinder())


class UboServiceThread(threading.Thread):
    def __init__(
        self,
        path: Path,
        *,
        allowed_service_ids: Sequence[str] | None = [],
    ) -> None:
        self.path = path
        self.allowed_service_ids = allowed_service_ids

        super().__init__(name=path.name)
        self.service_uid = self.path.as_posix().replace('.', '_')
        self.label = '<NOT SET>'
        self.should_auto_restart = False
        self.is_enabled = False

        self.module = None
        self.is_started = False
        self.has_reducer = False

        self._reducer_barrier = None
        self.subscriptions: Subscriptions = []

    def set_reducer_barrier(
        self,
        reducer_barrier: threading.Barrier,
    ) -> None:
        self._reducer_barrier = reducer_barrier

    def register_reducer(self, reducer: ReducerType) -> None:
        if self.has_reducer:
            msg = '`register_reducer` can only be called once per service'
            raise RuntimeError(msg)

        self.has_reducer = True
        from ubo_app.store.main import root_reducer_id, store

        logger.debug(
            'Registering ubo service reducer',
            extra={
                'service_id': self.service_id,
                'label': self.label,
                'reducer': f'{reducer.__module__}.{reducer.__name__}',
            },
        )

        store.dispatch(
            CombineReducerRegisterAction(
                combine_reducers_id=root_reducer_id,
                key=self.service_id,
                reducer=reducer,
            ),
        )

        self._wait_for_reducers()

    def _wait_for_reducers(self) -> None:
        if self._reducer_barrier:
            with contextlib.suppress(threading.BrokenBarrierError):
                self._reducer_barrier.wait()

    def register(  # noqa: PLR0913
        self,
        *,
        service_id: str,
        label: str,
        setup: SetupFunction,
        binary_path: str | None = None,
        binary_env_provider: Callable[[], dict[str, str]] | None = None,
        is_enabled: bool = True,
        should_auto_restart: bool = False,
    ) -> None:
        if (
            service_id in DISABLED_SERVICES
            or (ENABLED_SERVICES and service_id not in ENABLED_SERVICES)
            or (self.allowed_service_ids and service_id not in self.allowed_service_ids)
        ):
            logger.debug(
                'Skipping disabled ubo service',
                extra={
                    'service_id': service_id,
                    'label': label,
                    'enabled services': ENABLED_SERVICES,
                    'disabled services': DISABLED_SERVICES,
                    'allowed service ids': self.allowed_service_ids,
                },
            )
            msg = f'Service {service_id} is disabled'
            raise DisabledServiceError(msg)

        self.label = label
        self.service_id = service_id
        self.setup = setup
        self.binary_path = binary_path
        self.binary_env_provider = binary_env_provider
        self.is_enabled = is_enabled
        self.should_auto_restart = should_auto_restart

        logger.debug(
            'Ubo service registered!',
            extra={
                'service_id': self.service_id,
                'label': self.label,
            },
        )

    def initiate(self) -> None:
        try:
            if self.path.exists():
                self.spec = PathFinder.find_spec(
                    'ubo_handle',
                    [self.path.as_posix()],
                    None,
                )
                if not self.spec or not self.spec.origin:
                    return
                self.spec.name = f'{self.service_uid}:ubo_handle'
                self.spec.loader = UboModuleLoader(
                    'ubo_handle',
                    self.spec.origin,
                )
                self.module = importlib.util.module_from_spec(self.spec)

        except Exception:
            logger.exception('Error loading service', extra={'path': self.path})
            return

        if self.module and self.spec and self.spec.loader:
            SERVICES_BY_PATH[self.path] = self
            try:
                cast('UboServiceThread', self.module).register = self.register
                self.spec.loader.exec_module(self.module)
            except DisabledServiceError:
                del SERVICES_BY_PATH[self.path]
            except Exception:
                del SERVICES_BY_PATH[self.path]
                logger.exception('Error loading service', extra={'path': self.path})
            else:
                SERVICE_PATHS_BY_ID[self.service_id] = self.path

    def start(self) -> None:
        if not hasattr(self, 'setup'):
            return

        super().start()

    def stop(self) -> None:
        self.loop.call_soon_threadsafe(self.loop.create_task, self.shutdown())

    def run(self) -> None:  # noqa: C901, PLR0915
        from ubo_app.store.main import store

        self.loop = asyncio.new_event_loop()
        self.loop.set_exception_handler(loop_exception_handler)

        @store.autorun(
            lambda state: state.settings.services[self.service_id].log_level
            if self.service_id in state.settings.services
            else None,
        )
        def set_log_level(log_level: int | None) -> None:
            ThreadLevelFilter.set_thread_level(self.name, log_level)

        logger.info(
            'Starting service thread',
            extra={
                'thread_native_id': self.native_id,
                'service_label': self.label,
                'service_id': self.service_id,
            },
        )
        asyncio.set_event_loop(self.loop)

        async def setup_wrapper() -> None:  # noqa: C901
            try:
                result = None
                if len(inspect.signature(self.setup).parameters) == 0:
                    self._wait_for_reducers()
                    result = cast(
                        'Callable[[], SetupFunctionReturnType]',
                        self.setup,
                    )()
                elif len(inspect.signature(self.setup).parameters) == 1:
                    result = cast(
                        'Callable[[ReducerRegistrar], SetupFunctionReturnType]',
                        self.setup,
                    )(self.register_reducer)

                if asyncio.iscoroutine(result):
                    subscriptions = await result
                else:
                    subscriptions = result
                subscriptions = [*subscriptions] if subscriptions else []

                if self.binary_path is not None:
                    process_path = self.path / 'ubo-service' / self.binary_path
                    if process_path.exists():
                        process = await asyncio.subprocess.create_subprocess_exec(
                            process_path,
                            cwd=process_path.parent.parent,
                            env={
                                'GRPC_ADDRESS': GRPC_LISTEN_ADDRESS,
                                'GRPC_PORT': str(GRPC_LISTEN_PORT),
                                'PATH': os.environ.get('PATH', ''),
                                'UBO_DATA_PATH': DATA_PATH,
                                **(
                                    self.binary_env_provider()
                                    if self.binary_env_provider
                                    else {}
                                ),
                            },
                            start_new_session=True,
                        )

                        async def process_terminate() -> None:
                            if not process.returncode:
                                os.killpg(process.pid, signal.SIGTERM)
                                try:
                                    await asyncio.wait_for(
                                        process.wait(),
                                        timeout=SERVICES_LOOP_GRACE_PERIOD,
                                    )
                                except TimeoutError:
                                    os.killpg(process.pid, signal.SIGKILL)

                        subscriptions.append(process_terminate)

            except Exception:
                logger.exception(
                    'Error during setup',
                    extra={
                        'service_id': self.service_id,
                        'label': self.label,
                    },
                )
                raise

            self.is_started = True

            if subscriptions:
                self.subscriptions = subscriptions

            store.dispatch(
                SettingsServiceSetStatusAction(
                    service_id=self.service_id,
                    is_active=True,
                ),
            )

            del self.setup_task

        self.setup_task = setup_wrapper()
        self.loop.create_task(self.setup_task, name=f'Setup task for {self.label}')

        try:
            self.loop.run_forever()
        except Exception:
            logger.exception(
                'Ubo service thread ran into an error and stopped',
                extra={
                    'thread_native_id': self.native_id,
                    'service_label': self.label,
                    'service_id': self.service_id,
                },
            )
            self.kill()
            raise
        else:
            logger.info(
                'Ubo service thread stopped gracefully',
                extra={
                    'thread_native_id': self.native_id,
                    'service_label': self.label,
                    'service_id': self.service_id,
                },
            )
        finally:
            ThreadLevelFilter.set_thread_level(self.name, None)
            store.dispatch(
                SettingsServiceSetStatusAction(
                    service_id=self.service_id,
                    is_active=False,
                ),
            )

    def __repr__(self) -> str:
        return (
            f'<UboServiceThread id={self.service_id} '
            f'label={self.label} name={self.name} is_alive={self.is_alive()}>'
        )

    def run_coroutine(
        self,
        coroutine: Coroutine,
        callback: TaskCreatorCallback | None = None,
        name: str | None = None,
    ) -> asyncio.Handle:
        def task_wrapper(stack: str) -> None:
            task = self.loop.create_task(coroutine, name=name)
            if DEBUG_TASKS:
                STACKS[task] = stack
            if callback:
                callback(task)

        return self.loop.call_soon_threadsafe(
            task_wrapper,
            ''.join(traceback.format_stack()[:-3]) if DEBUG_TASKS else '',
        )

    async def shutdown(self) -> None:
        from ubo_app.logger import logger

        logger.debug(
            'Shutting down service thread',
            extra={
                'thread_native_id': self.native_id,
                'service_label': self.label,
                'service_id': self.service_id,
            },
        )

        await self._clean_subscriptions()
        await self._clean_remaining_tasks()
        self._unregister_reducer()
        self._cleanup()

    async def _clean_subscriptions(self) -> None:
        if not hasattr(self, 'subscriptions'):
            return
        subscriptions = self.subscriptions
        del self.subscriptions
        tasks: list[Coroutine] = []
        for unsubscribe in subscriptions:
            try:
                result = unsubscribe()
                if asyncio.iscoroutine(result):
                    tasks.append(result)
            except Exception:
                logger.exception(
                    'Error during cleanup',
                    extra={
                        'service_id': self.service_id,
                        'service_label': self.label,
                        'cleanup_callback': unsubscribe,
                    },
                )
        with contextlib.suppress(TimeoutError):
            await asyncio.wait_for(
                asyncio.gather(*tasks, return_exceptions=True),
                timeout=SERVICES_LOOP_GRACE_PERIOD,
            )

    async def _clean_remaining_tasks(self) -> None:
        if not self.loop.is_running():
            return
        while tasks := [
            task
            for task in asyncio.all_tasks(self.loop)
            if task is not asyncio.current_task(self.loop)
            and task.cancelling() == 0
            and not task.done()
        ]:
            logger.info(
                'Waiting for tasks to finish',
                extra={
                    'tasks': tasks,
                    'thread_': self,
                    'service_id': self.service_id,
                    'service_label': self.label,
                },
            )
            with contextlib.suppress(BaseException):
                await asyncio.wait_for(
                    asyncio.gather(
                        *tasks,
                        return_exceptions=True,
                    ),
                    timeout=SERVICES_LOOP_GRACE_PERIOD,
                )
            await asyncio.sleep(0.1)
        self.loop.stop()

    def _unregister_reducer(self) -> None:
        if self.has_reducer:
            from ubo_app.store.main import root_reducer_id, store

            store.dispatch(
                CombineReducerUnregisterAction(
                    combine_reducers_id=root_reducer_id,
                    key=self.service_id,
                ),
            )
            self.has_reducer = False

    def _cleanup(self) -> None:
        from ubo_app.utils import bus_provider

        if self in bus_provider.system_buses:
            del bus_provider.system_buses[self]
        if self in bus_provider.user_buses:
            del bus_provider.user_buses[self]

        for name in list(sys.modules):
            if name.startswith(f'{self.service_uid}:'):
                del sys.modules[name]

        if self.path in SERVICES_BY_PATH:
            del SERVICES_BY_PATH[self.path]

        del self.module

    def kill(self) -> None:
        if self.ident is None:
            return
        if self.loop.is_running():
            self.loop.create_task(self.shutdown())
        else:
            asyncio.new_event_loop().run_until_complete(self.shutdown())
        if not self.is_alive():
            return
        tid = ctypes.c_long(self.ident)
        res = ctypes.pythonapi.PyThreadState_SetAsyncExc(
            tid,
            ctypes.py_object(SystemExit),
        )
        if res > 1:
            ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, None)
            msg = 'Failed to stop thread'
            raise RuntimeError(msg)


async def start(event: SettingsStartServiceEvent) -> None:
    if event.service_id in SERVICE_PATHS_BY_ID and (
        SERVICE_PATHS_BY_ID[event.service_id] not in SERVICES_BY_PATH
        or not SERVICES_BY_PATH[SERVICE_PATHS_BY_ID[event.service_id]].is_started
    ):
        await asyncio.sleep(event.delay)
        if SERVICE_PATHS_BY_ID[event.service_id] not in SERVICES_BY_PATH:
            service = UboServiceThread(SERVICE_PATHS_BY_ID[event.service_id])
            service.initiate()
        SERVICES_BY_PATH[SERVICE_PATHS_BY_ID[event.service_id]].start()


async def stop(event: SettingsStopServiceEvent) -> None:
    if (
        event.service_id in SERVICE_PATHS_BY_ID
        and SERVICE_PATHS_BY_ID[event.service_id] in SERVICES_BY_PATH
        and SERVICES_BY_PATH[SERVICE_PATHS_BY_ID[event.service_id]].is_alive()
    ):
        service = SERVICES_BY_PATH[SERVICE_PATHS_BY_ID[event.service_id]]
        logger.debug(
            'Stopping service',
            extra={
                'service_id': event.service_id,
                'label': service.label,
            },
        )
        service.stop()
        await asyncio.to_thread(service.join, timeout=3)
        if service.is_alive():
            service.kill()
            await asyncio.to_thread(service.join, timeout=3)

        import gc

        gc.collect()


def _report_successful_reducer_registration(services: list[UboServiceThread]) -> None:
    logger.info(
        'All reducers registered successfully',
        extra={
            'service_ids': [service.service_id for service in services],
        },
    )


def stop_services(
    service_ids: Sequence[str] | None = None,
) -> None:
    for service in list(SERVICES_BY_PATH.values()):
        if service_ids and service.service_id not in service_ids:
            continue
        service.stop()


def load_services(
    service_ids: Sequence[str] | None = None,
    gap_duration: float = 0,
) -> None:
    from ubo_app.store.main import store
    from ubo_app.utils.persistent_store import read_from_persistent_store

    for services_directory_path in [
        ROOT_PATH.joinpath('services').as_posix(),
        *SERVICES_PATH,
    ]:
        directory_path = Path(services_directory_path).absolute()
        if directory_path.is_dir():
            for service_path in sorted(directory_path.iterdir()):
                if not service_path.is_dir() or service_path in SERVICES_BY_PATH.copy():
                    continue
                service = UboServiceThread(
                    service_path.absolute(),
                    allowed_service_ids=service_ids,
                )
                service.initiate()

    store.subscribe_event(SettingsStartServiceEvent, start)
    store.subscribe_event(SettingsStopServiceEvent, stop)

    services = {
        service.service_id: ServiceState(
            id=service.service_id,
            label=service.label,
            is_active=service.is_alive(),
            is_enabled=service.is_enabled,
            log_level=logging.INFO,
            should_auto_restart=service.should_auto_restart,
        )
        for service in SERVICES_BY_PATH.values()
    }

    for service in read_from_persistent_store(
        'services',
        default={},
    ):
        if service['id'] in SERVICE_PATHS_BY_ID:
            services[service['id']] = replace(
                services[service['id']],
                is_enabled=service.get(
                    'is_enabled',
                    services[service['id']].is_enabled,
                ),
                log_level=service.get('log_level', logging.INFO),
                should_auto_restart=service.get(
                    'should_auto_restart',
                    services[service['id']].should_auto_restart,
                ),
            )

    to_run_services = [
        service
        for service in SERVICES_BY_PATH.values()
        if service.is_enabled and (not service_ids or service.service_id in service_ids)
    ]

    reducer_barrier = threading.Barrier(
        len(to_run_services),
        action=lambda: _report_successful_reducer_registration(to_run_services),
        timeout=10,
    )

    for service in to_run_services:
        service.set_reducer_barrier(reducer_barrier)

    store.dispatch(
        SettingsSetServicesAction(services=services, gap_duration=gap_duration),
    )


__import__('ubo_app._ignore_this_in_back_track')



"""Deepgram engine interface."""

import re

from typing_extensions import override

from ubo_app.constants.assistant import (
    DEEPGRAM_API_KEY_PATTERN,
    DEEPGRAM_API_KEY_SECRET_ID,
)
from ubo_app.engines.abstraction.ai_provider_mixin import AIProviderMixin
from ubo_app.engines.abstraction.needs_setup_mixin import NeedsSetupMixin
from ubo_app.engines.abstraction.remote_mixin import RemoteMixin
from ubo_app.store.input.types import (
    InputFieldDescription,
    InputFieldType,
    QRCodeInputDescription,
    WebUIInputDescription,
)
from ubo_app.store.services.speech_synthesis import ReadableInformation
from ubo_app.utils import secrets
from ubo_app.utils.input import ubo_input


class DeepgramEngine(NeedsSetupMixin, AIProviderMixin, RemoteMixin):
    """Deepgram engine."""

    @property
    def name(self) -> str:
        """The internal name of the Deepgram engine."""
        return 'deepgram'

    @property
    def label(self) -> str:
        """The display label for the Deepgram engine."""
        return 'Deepgram'

    @property
    def not_setup_message(self) -> str:
        """Message shown when the Deepgram service API key is not set."""
        return 'Deepgram service API key is not set. You can set it in the settings.'

    @property
    @override
    def is_setup(self) -> bool:
        """Check if the Deepgram engine is set up."""
        service_account_info_string = secrets.read_secret(
            DEEPGRAM_API_KEY_SECRET_ID,
        )
        return (
            bool(service_account_info_string)
            and re.match(
                DEEPGRAM_API_KEY_PATTERN,
                service_account_info_string,
            )
            is not None
        )

    async def _setup(self) -> None:
        _, result = await ubo_input(
            title='Deepgram API Key',
            prompt='Enter your Deepgram API key.',
            descriptions=[
                WebUIInputDescription(
                    fields=[
                        InputFieldDescription(
                            name='api_key',
                            type=InputFieldType.TEXT,
                            label='API key',
                            description='Enter your Deepgram API key',
                            required=True,
                            pattern=DEEPGRAM_API_KEY_PATTERN,
                        ),
                    ],
                ),
                QRCodeInputDescription(
                    title='Deepgram API Key',
                    instructions=ReadableInformation(
                        text='Convert your Deepgram API key to a QR code and hold it '
                        'in front of the camera to scan it.',
                        picovoice_text='Convert your Deepgram API key to a '
                        '{QR|K Y UW AA R} code and hold it in front of the camera '
                        'to scan it.',
                    ),
                    pattern=r'(?P<api_key>' + DEEPGRAM_API_KEY_PATTERN + ')',
                ),
            ],
        )
        if result is None:
            return

        api_key = result.data.get('api_key')
        if not api_key:
            return

        secrets.write_secret(
            key=DEEPGRAM_API_KEY_SECRET_ID,
            value=api_key,
        )



# ruff: noqa: D100, D101
from __future__ import annotations

from dataclasses import field
from typing import TYPE_CHECKING

from immutable import Immutable
from redux import BaseAction

from ubo_app.store.core.types import service_default_factory

if TYPE_CHECKING:
    from collections.abc import Sequence


class IconState(Immutable):
    symbol: str
    color: str
    priority: int
    service_id: str
    id: str | None


class StatusIconsState(Immutable):
    icons: Sequence[IconState]


class StatusIconsAction(BaseAction): ...


class StatusIconsRegisterAction(StatusIconsAction):
    icon: str
    color: str = 'white'
    priority: int = 0
    id: str | None = None
    service: str | None = field(default_factory=service_default_factory)



# ruff: noqa: D100, D101
from __future__ import annotations

from enum import StrEnum


class NetState(StrEnum):
    CONNECTED = 'Connected'
    DISCONNECTED = 'Disconnected'
    PENDING = 'Pending'
    NEEDS_ATTENTION = 'Needs Attention'
    UNKNOWN = 'Unknown'



# ruff: noqa: D100, D101, D103
from __future__ import annotations

from dataclasses import field
from enum import StrEnum
from typing import TYPE_CHECKING

from immutable import Immutable
from redux import BaseAction, BaseEvent

from ubo_app.utils.service import ServiceUnavailableError

if TYPE_CHECKING:
    from collections.abc import Sequence

    from ubo_gui.menu.types import Item, Menu

    from ubo_app.store.ubo_actions import BasicType


class SettingsCategory(StrEnum):
    NETWORK = 'Network'
    REMOTE = 'Remote'
    SYSTEM = 'System'
    HARDWARE = 'Hardware'
    ASSISTANT = 'Assistant'
    DOCKER = 'Docker'
    ACCESSIBILITY = 'Accessibility'


class MenuScrollDirection(StrEnum):
    UP = 'up'
    DOWN = 'down'


SETTINGS_ICONS = {
    SettingsCategory.NETWORK: '󰛳',
    SettingsCategory.REMOTE: '󰑔',
    SettingsCategory.SYSTEM: '󰒔',
    SettingsCategory.HARDWARE: '',
    SettingsCategory.ASSISTANT: '󰚩',
    SettingsCategory.DOCKER: '󰡨',
    SettingsCategory.ACCESSIBILITY: '󰙋',
}


class MainAction(BaseAction): ...


class UpdateLightDMState(MainAction):
    is_active: bool
    is_enable: bool


def service_default_factory() -> str | None:
    from ubo_app.utils.service import get_service

    try:
        return get_service().service_id
    except ServiceUnavailableError:
        return None


class RegisterAppAction(MainAction):
    menu_item: Item
    service: str | None = field(default_factory=service_default_factory)
    key: str | None = None


class RegisterRegularAppAction(RegisterAppAction):
    priority: int | None = None


class DeregisterRegularAppAction(MainAction):
    service: str | None = field(default_factory=service_default_factory)
    key: str | None = None


class RegisterSettingAppAction(RegisterAppAction):
    category: SettingsCategory
    priority: int | None = None


class PowerAction(MainAction): ...


class PowerOffAction(PowerAction): ...


class RebootAction(PowerAction): ...


class SetMenuPathAction(MainAction):
    path: Sequence[str]
    depth: int


class SetAreEnclosuresVisibleAction(MainAction):
    is_header_visible: bool = True
    is_footer_visible: bool = True


class MenuAction(MainAction): ...


class MenuGoBackAction(MenuAction): ...


class MenuGoHomeAction(MenuAction): ...


class MenuChooseByIconAction(MenuAction):
    icon: str


class MenuChooseByLabelAction(MenuAction):
    label: str


class MenuChooseByIndexAction(MenuAction):
    index: int


class MenuScrollAction(MenuAction):
    direction: MenuScrollDirection


class OpenApplicationAction(MainAction):
    application_id: str
    initialization_args: tuple[BasicType, ...] = ()
    initialization_kwargs: dict[
        str,
        BasicType | tuple[BasicType, ...] | list[BasicType],
    ] = field(default_factory=dict)


class CloseApplicationAction(MainAction):
    application_instance_id: str


class MainEvent(BaseEvent): ...


class InitEvent(MainEvent): ...


class MenuEvent(MainEvent): ...


class MenuGoBackEvent(MenuEvent): ...


class MenuGoHomeEvent(MenuEvent): ...


class MenuChooseByIconEvent(MenuEvent):
    icon: str


class MenuChooseByLabelEvent(MenuEvent):
    label: str


class MenuChooseByIndexEvent(MenuEvent):
    index: int


class MenuScrollEvent(MenuEvent):
    direction: MenuScrollDirection


class OpenApplicationEvent(MainEvent):
    application_id: str
    initialization_args: tuple[BasicType, ...] = ()
    initialization_kwargs: dict[
        str,
        BasicType | tuple[BasicType, ...] | list[BasicType],
    ] = field(default_factory=dict)


class CloseApplicationEvent(MainEvent):
    application_instance_id: str


class PowerEvent(MainEvent): ...


class PowerOffEvent(PowerEvent): ...


class RebootEvent(PowerEvent): ...


class ScreenshotEvent(MainEvent):
    """Event for taking a screenshot."""


class SnapshotEvent(MainEvent):
    """Event for taking a snapshot of the store."""


class ToggleRecordingAction(MainAction):
    """Action for toggling recording."""


class StoreRecordedSequenceEvent(MainEvent):
    """Event for storing a recorded sequence."""

    recorded_sequence: list[BaseAction]


class ReplayRecordedSequenceAction(MainAction):
    """Action for replaying a recorded sequence."""


class ReplayRecordedSequenceEvent(MainEvent):
    """Event for replaying a recorded sequence."""


class ReportReplayingDoneAction(MainAction):
    """Action for reporting that replaying is done."""


class MainState(Immutable):
    menu: Menu | None = None
    path: Sequence[str] = field(default_factory=list)
    depth: int = 0
    is_header_visible: bool = True
    is_footer_visible: bool = True
    apps_items_priorities: dict[str, int] = field(default_factory=dict)
    settings_items_priorities: dict[str, int] = field(default_factory=dict)
    is_recording: bool = False
    is_replaying: bool = False
    recorded_sequence: list[BaseAction] = field(default_factory=list)



"""Base class for third-party engines."""

import abc


class EngineMixin(abc.ABC):
    """Base class for third-party engines."""

    def __init__(self, *, label: str | None = None) -> None:
        """Initialize the EngineMixin."""
        self._instance_label = label

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """The name of the engine."""

    @property
    @abc.abstractmethod
    def label(self) -> str:
        """The label of the engine."""

    @property
    def instance_label(self) -> str:
        """The label of the engine instance."""
        return self.label if self._instance_label is None else self._instance_label



# ruff: noqa: D100, D101, D102, D107
from __future__ import annotations

import functools
import math
import weakref
from dataclasses import replace
from typing import TYPE_CHECKING

from kivy.clock import Clock, mainthread
from kivy.properties import StringProperty
from ubo_gui.app import UboApp
from ubo_gui.menu.stack_item import StackApplicationItem
from ubo_gui.menu.types import HeadlessMenu
from ubo_gui.notification import NotificationWidget
from ubo_gui.page import PAGE_MAX_ITEMS

from ubo_app.colors import INFO_COLOR
from ubo_app.logger import logger
from ubo_app.menu_app.notification_info import NotificationInfo
from ubo_app.store.core.types import CloseApplicationAction
from ubo_app.store.main import store
from ubo_app.store.services.notifications import (
    Notification,
    NotificationActionItem,
    NotificationApplicationItem,
    NotificationDisplayType,
    NotificationsClearAction,
    NotificationsClearEvent,
    NotificationsDisplayEvent,
)
from ubo_app.store.services.speech_synthesis import SpeechSynthesisReadTextAction
from ubo_app.utils.gui import UboPageWidget

if TYPE_CHECKING:
    from collections.abc import Callable

    from kivy._clock import ClockEvent
    from ubo_gui.menu.menu_widget import MenuWidget
    from ubo_gui.menu.types import Menu, PageWidget


class NotificationReference:
    def __init__(self, notification: Notification) -> None:
        self.value = notification
        self.dismiss_on_close = notification.dismiss_on_close
        self.is_initialized = False
        self.flash_event: ClockEvent | None = None


class UboNotificationWidget(NotificationWidget, UboPageWidget):
    """renders a notification."""

    notification_id: str = StringProperty()


class MenuNotificationHandler(UboApp):
    menu_widget: MenuWidget

    @mainthread
    def display_notification(  # noqa: C901
        self,
        event: NotificationsDisplayEvent,
    ) -> None:
        if (
            event.notification.id
            and any(
                isinstance(stack_item, StackApplicationItem)
                and isinstance(stack_item.application, UboNotificationWidget)
                and stack_item.application.notification_id == event.notification.id
                for stack_item in self.menu_widget.stack
            )
        ) or (
            event.notification.display_type is NotificationDisplayType.BACKGROUND
            and event.index is None
        ):
            return

        subscriptions = []

        notification = NotificationReference(event.notification)
        is_closed = False

        logger.debug('Opening notification %s', notification.value.id)

        @mainthread
        def close(_: object = None) -> None:
            nonlocal is_closed
            logger.debug(
                'Closing notification %s',
                notification.value.id,
                extra={'is_closed': is_closed},
            )
            if is_closed:
                return
            is_closed = True
            for unsubscribe in subscriptions:
                unsubscribe()
            notification_application.unbind(on_close=close)
            store.dispatch(
                CloseApplicationAction(
                    application_instance_id=notification_application.id,
                ),
            )
            if notification.dismiss_on_close:
                store.dispatch(
                    NotificationsClearAction(notification=notification.value),
                )
            if notification.value.on_close:
                notification.value.on_close()

        def clear_notification(event: NotificationsClearEvent) -> None:
            if event.notification == notification.value:
                close()

        _self = weakref.ref(self)

        def renew_notification(event: NotificationsDisplayEvent) -> None:
            logger.verbose('Renewing notification', extra={'notification': event})
            self = _self()
            if self is None:
                return
            if event.notification.id == notification.value.id:
                notification.value = event.notification
                notification.dismiss_on_close = event.notification.dismiss_on_close
                self._update_notification_widget(
                    notification_application,
                    event,
                    notification,
                    close,
                )

                if notification.flash_event:
                    notification.flash_event.cancel()
                    notification.flash_event = None

                if (
                    event.notification.display_type is NotificationDisplayType.FLASH
                    and event.index is None
                ):
                    notification.flash_event = Clock.schedule_once(
                        close,
                        notification.value.flash_time,
                    )

            if event.notification.extra_information and (
                not notification.is_initialized
                or event.notification.id is None
                or event.notification.id != notification.value.id
                or not notification.value.extra_information
                or event.notification.extra_information
                != notification.value.extra_information
            ):
                notification.is_initialized = True
                store.dispatch(
                    SpeechSynthesisReadTextAction(
                        information=event.notification.extra_information,
                    ),
                )

        notification_application = UboNotificationWidget(
            notification_id=notification.value.id,
            items=[None] * PAGE_MAX_ITEMS,
        )

        notification_application.bind(on_close=close)

        subscriptions.append(
            store.subscribe_event(
                NotificationsClearEvent,
                clear_notification,
            ),
        )
        if notification.value.id is not None:
            subscriptions.append(
                store.subscribe_event(
                    NotificationsDisplayEvent,
                    renew_notification,
                ),
            )

        renew_notification(event)

        self.menu_widget.open_application(notification_application)

    def _notification_items(  # noqa: C901
        self,
        notification: NotificationReference,
        close: Callable[[], None],
    ) -> list[NotificationActionItem | NotificationApplicationItem | None]:
        def dismiss(_: object = None) -> None:
            close()
            if not notification.dismiss_on_close:
                store.dispatch(
                    NotificationsClearAction(notification=notification.value),
                )

        def run_notification_action(
            action: NotificationActionItem,
        ) -> Menu | Callable[[], Menu] | type[PageWidget] | PageWidget | None:
            result = action.action()
            if action.close_notification:
                if action.dismiss_notification:
                    dismiss()
                else:
                    close()
            return result

        top_items: list[
            NotificationActionItem | NotificationApplicationItem | None
        ] = []
        bottom_items: list[
            NotificationActionItem | NotificationApplicationItem | None
        ] = []

        if notification.value.extra_information:
            text = notification.value.extra_information.text

            top_items.append(
                NotificationActionItem(
                    key='info',
                    icon='󰋼',
                    action=lambda: NotificationInfo(text=text),
                    label='',
                    is_short=True,
                    background_color=INFO_COLOR,
                ),
            )

        def get_application_runner(
            action: NotificationApplicationItem,
        ) -> Callable[[], PageWidget | type[PageWidget]]:
            def run_application() -> PageWidget | type[PageWidget]:
                if callable(action.application) and not isinstance(
                    action.application,
                    type,
                ):
                    return action.application()
                return action.application

            return run_application

        if notification.value.show_dismiss_action:
            bottom_items.append(
                NotificationActionItem(
                    key='dismiss',
                    icon='',
                    action=dismiss,
                    label='',
                    is_short=True,
                    background_color='#C0C0C0',
                ),
            )

        actions_quantity = (
            len(top_items) + len(notification.value.actions) + len(bottom_items)
        )

        def convert_action(
            action: NotificationActionItem | NotificationApplicationItem,
        ) -> NotificationActionItem:
            return (
                replace(
                    action,
                    is_short=True,
                    action=(
                        action_ := functools.partial(
                            run_notification_action,
                            action,
                        ),
                        setattr(
                            action_,
                            '_is_default_action_of_ubo_dispatch_item',
                            True,
                        ),
                    )[0],
                )
                if isinstance(action, NotificationActionItem)
                else NotificationActionItem(
                    action=get_application_runner(action),
                    background_color=action.background_color,
                    close_notification=action.close_notification,
                    color=action.color,
                    dismiss_notification=action.dismiss_notification,
                    icon=action.icon,
                    is_short=True,
                    key=action.key,
                    label=action.label,
                    opacity=action.opacity,
                    progress=action.progress,
                )
            )

        if actions_quantity > PAGE_MAX_ITEMS:

            def open_options() -> HeadlessMenu:
                return HeadlessMenu(
                    title=notification.value.icon + ' Select',
                    items=[
                        convert_action(action) for action in notification.value.actions
                    ],
                )

            return (
                top_items
                + [
                    convert_action(action)
                    for action in notification.value.actions[
                        : PAGE_MAX_ITEMS - len(top_items) - len(bottom_items) - 1
                    ]
                ]
                + [
                    NotificationActionItem(
                        key='all',
                        icon='󰍜',
                        action=open_options,
                        is_short=True,
                    ),
                ]
                + bottom_items
            )

        items = (
            top_items
            + [convert_action(action) for action in notification.value.actions]
            + bottom_items
        )
        return [None] * (PAGE_MAX_ITEMS - len(items)) + items

    @mainthread
    def _update_notification_widget(
        self,
        notification_application: UboNotificationWidget,
        event: NotificationsDisplayEvent,
        notification: NotificationReference,
        close: Callable[[], None],
    ) -> None:
        notification_application.notification_title = notification.value.title
        notification_application.content = notification.value.content
        notification_application.icon = notification.value.icon + (
            f'[size=20dp] {notification.value.progress:05.1%}[/size]'
            if notification.value.progress is not None
            and not math.isnan(notification.value.progress)
            else ''
        )
        notification_application.color = notification.value.color
        notification_application.items = self._notification_items(
            notification,
            close,
        )
        notification_application.title = (
            f'Notification ({event.index + 1}/{event.count})'
            if event.index is not None
            else ' '
        )



"""Initialize the infrared service."""

from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

from ubo_gui.menu.types import HeadlessMenu, Item, SubMenuItem

from ubo_app.logger import logger
from ubo_app.store.core.types import RegisterSettingAppAction, SettingsCategory
from ubo_app.store.main import store
from ubo_app.store.services.infrared import (
    InfraredHandleReceivedCodeAction,
    InfraredSendCodeEvent,
    InfraredSetShouldPropagateAction,
    InfraredSetShouldReceiveAction,
)
from ubo_app.store.ubo_actions import UboDispatchItem
from ubo_app.utils.async_ import create_task
from ubo_app.utils.gui import SELECTED_ITEM_PARAMETERS, UNSELECTED_ITEM_PARAMETERS
from ubo_app.utils.persistent_store import register_persistent_store
from ubo_app.utils.server import send_command

if TYPE_CHECKING:
    from ubo_app.utils.types import Subscriptions


@store.with_state(lambda state: state.infrared.should_receive_keypad_actions)
def _should_receive_keypad_actions(value: bool) -> bool:  # noqa: FBT001
    return value


ir_ctl_lock = asyncio.Lock()
ir_commands_queue = asyncio.Queue()


async def _send_code(action: InfraredSendCodeEvent) -> None:
    await ir_commands_queue.put(action)
    async with ir_ctl_lock:
        action = await ir_commands_queue.get()
        logger.debug(
            'Sending infrared code.',
            extra={'protocol': action.protocol, 'scancode': action.scancode},
        )

        process = await asyncio.create_subprocess_exec(
            'ir-ctl',
            '-S',
            f'{action.protocol}:{action.scancode}',
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        await asyncio.wait_for(process.wait(), timeout=1)
        if process.returncode is None:
            process.kill()
            msg = 'Infrared: Failed to send code, process killed due to timeout.'
            raise RuntimeError(msg)
        await asyncio.sleep(0.25)


async def _wait_for_ir_code() -> None:
    """Wait for IR codes from the system manager and dispatch them to the store."""
    while _should_receive_keypad_actions():
        try:
            async for response in await send_command(
                'infrared',
                'receive',
                has_output_stream=True,
            ):
                if response == 'nocode':
                    break
                protocol, scancode = response.split(':')
                logger.info(
                    'Received IR code from system manager',
                    extra={'protocol': protocol, 'scancode': scancode},
                )
                store.dispatch(
                    InfraredHandleReceivedCodeAction(
                        protocol=protocol,
                        scancode=scancode,
                    ),
                )
        except Exception:
            logger.exception('Failed to send infrared receive command')
            raise


def init_service() -> Subscriptions:
    """Initialize the infrared service."""
    ir_code_task: asyncio.Handle | None = None

    @store.autorun(lambda state: state.infrared.should_receive_keypad_actions)
    async def run_monitor_ir(value: bool) -> None:  # noqa: FBT001
        nonlocal ir_code_task
        if value:
            await send_command('infrared', 'start')
            ir_code_task = create_task(_wait_for_ir_code())
        else:
            await send_command('infrared', 'stop')
            if ir_code_task is not None:
                ir_code_task.cancel()

    register_persistent_store(
        'infrared_state:should_propagate_keypad_actions',
        lambda state: state.infrared.should_propagate_keypad_actions,
    )
    register_persistent_store(
        'infrared_state:should_receive_keypad_actions',
        lambda state: state.infrared.should_receive_keypad_actions,
    )

    @store.autorun(
        lambda state: (
            state.infrared.should_propagate_keypad_actions,
            state.infrared.should_receive_keypad_actions,
        ),
    )
    def menu_items(data: tuple[bool, bool]) -> list[Item]:
        should_propagate_keypad_actions, should_receive_keypad_actions = data
        return [
            UboDispatchItem(
                key='propagate_keys',
                label='Propagate Keys',
                store_action=InfraredSetShouldPropagateAction(
                    should_propagate=not should_propagate_keypad_actions,
                ),
                **(
                    SELECTED_ITEM_PARAMETERS
                    if should_propagate_keypad_actions
                    else UNSELECTED_ITEM_PARAMETERS
                ),
            ),
            UboDispatchItem(
                key='receive_keys',
                label='Receive Keys',
                store_action=InfraredSetShouldReceiveAction(
                    should_receive=not should_receive_keypad_actions,
                ),
                **(
                    SELECTED_ITEM_PARAMETERS
                    if should_receive_keypad_actions
                    else UNSELECTED_ITEM_PARAMETERS
                ),
            ),
        ]

    store.dispatch(
        RegisterSettingAppAction(
            key='infrared',
            category=SettingsCategory.HARDWARE,
            menu_item=SubMenuItem(
                label='Infrared',
                icon='󰻅',
                sub_menu=HeadlessMenu(
                    title='󰻅Infrared',
                    items=menu_items,
                ),
            ),
        ),
    )

    return [
        store.subscribe_event(InfraredSendCodeEvent, _send_code),
    ]



"""Ollama on-premise (remote) assistant engine module."""

from __future__ import annotations

import asyncio
import math
import re

import ollama
from typing_extensions import override

from ubo_app.colors import SUCCESS_COLOR, WARNING_COLOR
from ubo_app.constants.assistant import (
    OLLAMA_ONPREM_SETUP_NOTIFICATION_ID,
    OLLAMA_ONPREM_URL_PATTERN,
    OLLAMA_ONPREM_URL_SECRET_ID,
)
from ubo_app.engines.abstraction.ai_provider_mixin import AIProviderMixin
from ubo_app.engines.abstraction.needs_setup_mixin import NeedsSetupMixin
from ubo_app.engines.abstraction.remote_mixin import RemoteMixin
from ubo_app.logger import logger
from ubo_app.store.input.types import (
    InputFieldDescription,
    InputFieldType,
    QRCodeInputDescription,
    WebUIInputDescription,
)
from ubo_app.store.main import store
from ubo_app.store.services.assistant import (
    AssistantLLMName,
    AssistantSetSelectedModelAction,
    AssistantUpdateProvidersAction,
)
from ubo_app.store.services.notifications import (
    Notification,
    NotificationActionItem,
    NotificationDisplayType,
    NotificationsAddAction,
)
from ubo_app.store.services.speech_synthesis import ReadableInformation
from ubo_app.utils import secrets
from ubo_app.utils.async_ import create_task
from ubo_app.utils.error_handlers import report_service_error
from ubo_app.utils.input import ubo_input


class OllamaOnPremEngine(NeedsSetupMixin, AIProviderMixin, RemoteMixin):
    """Ollama on-premise (remote) assistant engine."""

    @property
    def name(self) -> AssistantLLMName:
        """Returns the name identifier for the Ollama on-premise assistant."""
        return AssistantLLMName.OLLAMA_ONPREM

    @property
    def label(self) -> str:
        """Returns the display label for the Ollama on-premise assistant."""
        return 'Ollama (On-Prem)'

    @property
    def not_setup_message(self) -> str:
        """Returns the message shown when Ollama on-premise is not set up."""
        return 'Ollama on-premise URL is not set. Please configure it in settings.'

    @property
    @override
    @store.with_state(
        lambda state: (
            state.assistant.selected_models.get(
                AssistantLLMName.OLLAMA_ONPREM,
                '',
            ),
            secrets.read_secret(OLLAMA_ONPREM_URL_SECRET_ID) or '',
        ),
    )
    def is_setup(self, model_and_url: tuple[str, str]) -> bool:  # noqa: PLR0206
        """Check if the remote Ollama instance is accessible and has the model."""
        model, url = model_and_url
        if not url:
            return False

        try:
            client = ollama.Client(host=url)
            models = client.list().models
        except ConnectionError:
            return False
        except Exception:
            logger.exception(
                'Error checking remote Ollama setup',
                extra={'url': url},
            )
            return False
        else:
            return any(m.model == model for m in models)

    @store.with_state(
        lambda state: (
            state.assistant.selected_models.get(
                AssistantLLMName.OLLAMA_ONPREM,
                '',
            ),
            secrets.read_secret(OLLAMA_ONPREM_URL_SECRET_ID) or '',
        ),
    )
    def _download_model(self, model_and_url: tuple[str, str]) -> None:
        """Download the specified Ollama model on the remote instance."""
        model, url = model_and_url

        async def download_ollama_model() -> None:
            """Download Ollama model to remote instance."""
            client = ollama.AsyncClient(host=url)
            progress_notification = Notification(
                id=OLLAMA_ONPREM_SETUP_NOTIFICATION_ID,
                title='Ollama (On-Prem)',
                content=f'Downloading {model} model to remote instance',
                icon='󰇚',
                color=WARNING_COLOR,
                display_type=NotificationDisplayType.STICKY,
                progress=0,
                show_dismiss_action=False,
                dismiss_on_close=False,
                blink=False,
            )
            store.dispatch(
                NotificationsAddAction(notification=progress_notification),
            )

            try:
                async for response in await client.pull(model, stream=True):
                    store.dispatch(
                        NotificationsAddAction(
                            notification=progress_notification(
                                progress=(response.completed / response.total)
                                if response.completed is not None
                                and response.total is not None
                                else math.nan,
                            ),
                        ),
                    )
            except Exception:
                logger.exception(
                    'Assistant - Error downloading Ollama model to remote '
                    'instance',
                    extra={'model': model, 'url': url},
                )
                report_service_error()
            else:
                logger.info(
                    'Ollama on-prem model download complete, updating providers',
                    extra={'model': model, 'url': url},
                )
                store.dispatch(
                    NotificationsAddAction(
                        notification=progress_notification(
                            content=(
                                f'"{model}" downloaded successfully to '
                                'remote instance'
                            ),
                            icon='󰄬',
                            color=SUCCESS_COLOR,
                            display_type=NotificationDisplayType.FLASH,
                            progress=None,
                        ),
                    ),
                    AssistantSetSelectedModelAction(model=model),
                    AssistantUpdateProvidersAction(),
                )
            finally:
                self.event.set()

        create_task(download_ollama_model())

    @override
    async def _setup(self) -> None:
        """Set up the remote Ollama instance by getting the URL from user."""
        # First, check if we already have a URL configured
        existing_url = secrets.read_secret(OLLAMA_ONPREM_URL_SECRET_ID)

        if existing_url and re.match(
            OLLAMA_ONPREM_URL_PATTERN,
            existing_url,
        ):
            # URL exists, check if it's accessible and prompt for model
            # download
            try:
                client = ollama.Client(host=existing_url)
                client.list()  # Test connection

                # Connection successful, offer to download model
                if self.is_setup:
                    return

                self.event = asyncio.Event()
                store.dispatch(
                    NotificationsAddAction(
                        notification=Notification(
                            id=OLLAMA_ONPREM_SETUP_NOTIFICATION_ID,
                            title='Ollama On-Prem Setup',
                            content=(
                                'Connected to remote Ollama. Download a '
                                'model to continue.'
                            ),
                            color=WARNING_COLOR,
                            actions=[
                                NotificationActionItem(
                                    label='Download Model',
                                    icon='󰇚',
                                    action=self._download_model,
                                ),
                            ],
                        ),
                    ),
                )
                await self.event.wait()
            except ConnectionError:
                # Connection failed, prompt for new URL
                logger.debug(
                    'Failed to connect to existing Ollama URL',
                    extra={'url': existing_url},
                )
            except Exception:
                # Unexpected error, prompt for new URL
                logger.exception(
                    'Error checking existing Ollama URL',
                    extra={'url': existing_url},
                )
            else:
                # Connection successful and setup complete
                return

        # Prompt for URL configuration
        _, result = await ubo_input(
            title='Ollama On-Prem URL',
            prompt='Enter the URL of your remote Ollama instance.',
            descriptions=[
                WebUIInputDescription(
                    fields=[
                        InputFieldDescription(
                            name='url',
                            type=InputFieldType.TEXT,
                            label='Ollama URL',
                            description=(
                                'Enter the URL '
                                '(e.g., http://192.168.1.100:11434)'
                            ),
                            required=True,
                            pattern=OLLAMA_ONPREM_URL_PATTERN,
                        ),
                    ],
                ),
                QRCodeInputDescription(
                    title='Ollama On-Prem URL',
                    instructions=ReadableInformation(
                        text=(
                            'Convert your Ollama URL to a QR code and '
                            'hold it in front of the camera to scan it.'
                        ),
                        picovoice_text=(
                            'Convert your Ollama URL to a '
                            '{QR|K Y UW AA R} code and hold it in '
                            'front of the camera to scan it.'
                        ),
                    ),
                    pattern=r'(?P<url>' + OLLAMA_ONPREM_URL_PATTERN + ')',
                ),
            ],
        )

        url = result.data['url']

        # Validate connection to the remote instance
        try:
            client = ollama.Client(host=url)
            client.list()  # Test connection
        except Exception:
            logger.exception(
                'Failed to connect to remote Ollama instance',
                extra={'url': url},
            )
            store.dispatch(
                NotificationsAddAction(
                    notification=Notification(
                        id=OLLAMA_ONPREM_SETUP_NOTIFICATION_ID,
                        title='Ollama On-Prem Setup Failed',
                        content=(
                            f'Cannot connect to {url}. '
                            'Please verify the URL.'
                        ),
                        color=WARNING_COLOR,
                        display_type=NotificationDisplayType.FLASH,
                    ),
                ),
            )
            return

        # Save the URL
        secrets.write_secret(
            key=OLLAMA_ONPREM_URL_SECRET_ID,
            value=url,
        )

        # Prompt for model download
        if not self.is_setup:
            self.event = asyncio.Event()
            store.dispatch(
                NotificationsAddAction(
                    notification=Notification(
                        id=OLLAMA_ONPREM_SETUP_NOTIFICATION_ID,
                        title='Ollama On-Prem Setup',
                        content=(
                            'Connected successfully. Download a model to '
                            'continue.'
                        ),
                        color=WARNING_COLOR,
                        actions=[
                            NotificationActionItem(
                                label='Download Model',
                                icon='󰇚',
                                action=self._download_model,
                            ),
                        ],
                    ),
                ),
            )
            await self.event.wait()



# ruff: noqa: D100, D103
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ubo_handle import ReducerRegistrar, register


async def setup(register_reducer: ReducerRegistrar) -> None:
    from reducer import reducer

    register_reducer(reducer)

    from setup import init_service

    await init_service()


def binary_env_provider() -> dict[str, str]:
    from ubo_app.constants.assistant import (
        ASSEMBLYAI_API_KEY_SECRET_ID,
        CEREBRAS_API_KEY_SECRET_ID,
        DEEPGRAM_API_KEY_SECRET_ID,
        ELEVENLABS_API_KEY_SECRET_ID,
        ELEVENLABS_VOICE_ID,
        GOOGLE_API_KEY_SECRET_ID,
        GOOGLE_CLOUD_SERVICE_ACCOUNT_KEY_SECRET_ID,
        GROK_API_KEY_SECRET_ID,
        OLLAMA_ONPREM_URL_SECRET_ID,
        OPENAI_API_KEY_SECRET_ID,
        RIME_API_KEY_SECRET_ID,
    )

    return {
        'ASSEMBLYAI_API_KEY_SECRET_ID': ASSEMBLYAI_API_KEY_SECRET_ID,
        'GOOGLE_API_KEY_SECRET_ID': GOOGLE_API_KEY_SECRET_ID,
        'GOOGLE_CLOUD_SERVICE_ACCOUNT_KEY_SECRET_ID': (
            GOOGLE_CLOUD_SERVICE_ACCOUNT_KEY_SECRET_ID
        ),
        'OPENAI_API_KEY_SECRET_ID': OPENAI_API_KEY_SECRET_ID,
        'GROK_API_KEY_SECRET_ID': GROK_API_KEY_SECRET_ID,
        'CEREBRAS_API_KEY_SECRET_ID': CEREBRAS_API_KEY_SECRET_ID,
        'OLLAMA_ONPREM_URL_SECRET_ID': OLLAMA_ONPREM_URL_SECRET_ID,
        'ELEVENLABS_API_KEY_SECRET_ID': ELEVENLABS_API_KEY_SECRET_ID,
        'ELEVENLABS_VOICE_ID': ELEVENLABS_VOICE_ID,
        'DEEPGRAM_API_KEY_SECRET_ID': DEEPGRAM_API_KEY_SECRET_ID,
        'RIME_API_KEY_SECRET_ID': RIME_API_KEY_SECRET_ID,
    }


register(
    service_id='assistant',
    label='Assistant',
    setup=setup,
    binary_path='bin/ubo-assistant',
    binary_env_provider=binary_env_provider,
)



# ruff: noqa: D100, D101, D102, D107
from __future__ import annotations

import pathlib
from functools import cached_property
from typing import TYPE_CHECKING

from kivy.clock import Clock, mainthread
from kivy.lang.builder import Builder
from ubo_gui.gauge import GaugeWidget
from ubo_gui.menu.constants import PAGE_SIZE
from ubo_gui.menu.menu_widget import MenuPageWidget
from ubo_gui.volume import VolumeWidget

from ubo_app.store.main import store

if TYPE_CHECKING:
    from collections.abc import Sequence

    from ubo_gui.menu.types import Item


class HomePage(MenuPageWidget):
    def __init__(
        self: HomePage,
        items: Sequence[Item | None] = [],
        *args: object,
        **kwargs: object,
    ) -> None:
        super().__init__(
            [None, *items, None],
            *args,
            **kwargs,
            count=PAGE_SIZE,
            render_surroundings=False,
        )

        self.ids.central_column.add_widget(self.cpu_gauge)
        self.ids.central_column.add_widget(self.ram_gauge)

        self.volume_widget = VolumeWidget()
        self.ids.right_column.add_widget(self.volume_widget)

        store.autorun(lambda state: state.audio.playback_volume)(
            self._sync_output_volume,
        )

    @mainthread
    def _sync_output_volume(self: HomePage, selector_result: float) -> None:
        self.volume_widget.value = selector_result * 100

    @cached_property
    def cpu_gauge(self: HomePage) -> GaugeWidget:
        import psutil

        gauge = GaugeWidget(
            value=psutil.cpu_percent(percpu=False),
            fill_color='#24D636',
            label='CPU',
        )

        def set_value(_: float) -> None:
            gauge.value = psutil.cpu_percent(percpu=False)

        Clock.schedule_interval(set_value, 1)

        return gauge

    @cached_property
    def ram_gauge(self: HomePage) -> GaugeWidget:
        import psutil

        gauge = GaugeWidget(
            value=psutil.virtual_memory().percent,
            fill_color='#D68F24',
            label='RAM',
        )

        def set_value(_: float) -> None:
            gauge.value = psutil.virtual_memory().percent

        Clock.schedule_interval(set_value, 1)

        return gauge


Builder.load_file(
    pathlib.Path(__file__).parent.joinpath('home_page.kv').resolve().as_posix(),
)



"""provides a function to interact with system services."""

from __future__ import annotations

import subprocess

from ubo_app.logger import get_logger


def service_handler(service: str, command: str) -> str | None:
    """Interact with system services."""
    logger = get_logger('system-manager')
    if service in ('ssh', 'lightdm'):
        try:
            if command == 'start':
                subprocess.run(  # noqa: S603
                    ['/usr/bin/env', 'systemctl', 'start', service],
                    check=True,
                )
            elif command == 'stop':
                subprocess.run(  # noqa: S603
                    ['/usr/bin/env', 'systemctl', 'stop', service],
                    check=True,
                )
            elif command == 'enable':
                subprocess.run(  # noqa: S603
                    ['/usr/bin/env', 'systemctl', 'enable', service],
                    check=True,
                )
            elif command == 'disable':
                subprocess.run(  # noqa: S603
                    ['/usr/bin/env', 'systemctl', 'disable', service],
                    check=True,
                )
        except Exception:
            logger.exception('Failed to handle service command.')
        else:
            return 'done'
    return None



# ruff: noqa: D100
from __future__ import annotations

import os
from pathlib import Path

import dotenv

from ubo_app.logger import setup_loggers
from ubo_app.setup import setup
from ubo_app.utils import IS_RPI
from ubo_app.utils.error_handlers import setup_error_handling

dotenv.load_dotenv(Path(__file__).parent / '.dev.env')
dotenv.load_dotenv(Path(__file__).parent / '.env')


def main() -> None:
    """Instantiate the `MenuApp` and run it."""
    # `setup_logging` imports `ubo_gui` which imports `kivy`, so we need to set these
    # environment variables before calling `setup_logging`
    os.environ['KIVY_NO_CONFIG'] = '1'
    os.environ['KIVY_NO_FILELOG'] = '1'
    os.environ['KIVY_NO_CONSOLELOG'] = '1'
    # We want to have full control over the exit behavior
    os.environ['KCFG_KIVY_EXIT_ON_ESCAPE'] = '0'
    # Hardware FBO is needed to make sure the in memory buffer uses the GPU
    os.environ['KCFG_GRAPHICS_FBO'] = 'force-hardware'
    # Anti-aliasing is not needed since Kivy is not directly rendering on a display
    os.environ['KCFG_GRAPHICS_MULTISAMPLES'] = '0'
    # V-SYNC is not needed since Kivy is not directly rendering on a display
    os.environ['KCFG_GRAPHICS_VSYNC'] = '0'

    # `setup_logging` needs to be called before anything else to initialize the rotating
    # log files
    logger_cleanups = setup_loggers()

    # `setup_error_handling` needs to be called before anything else and after
    # `setup_logging`
    setup_error_handling()

    # This should be imported early to set the custom loader
    from ubo_app.service_thread import load_services, stop_services

    setup()

    import headless_kivy.config

    from ubo_app.constants import (
        BYTES_PER_PIXEL,
        DISABLE_GRPC,
        DISPLAY_BAUDRATE,
        HEIGHT,
        WIDTH,
    )
    from ubo_app.display import render_on_display

    headless_kivy.config.setup_headless_kivy(
        headless_kivy.config.SetupHeadlessConfig(
            bandwidth_limit=DISPLAY_BAUDRATE // BYTES_PER_PIXEL // 8 if IS_RPI else 0,
            bandwidth_limit_window=0.025,
            bandwidth_limit_overhead=1000,
            region_size=60,
            callback=render_on_display,
            flip_vertical=True,
            width=WIDTH,
            height=HEIGHT,
        ),
    )

    from ubo_app.logger import logger
    from ubo_app.menu_app.menu import MenuApp

    logger.info('----------------------Starting the app----------------------')

    if not DISABLE_GRPC:
        from ubo_app.rpc.server import serve as grpc_serve
        from ubo_app.service import worker_thread

        worker_thread.run_coroutine(grpc_serve())

    load_services()
    app = MenuApp()

    from kivy.clock import mainthread
    from redux import FinishAction, FinishEvent

    from ubo_app.store.main import store

    store.subscribe_event(FinishEvent, mainthread(app.stop))

    if not IS_RPI:
        from kivy.core.window import Window

        Window._win.set_always_on_top(True)  # noqa: SLF001

    try:
        app.run()
    except Exception:
        logger.exception('An error occurred while running the app.')

        store.dispatch(FinishAction())
    finally:
        from ubo_app.service_thread import SERVICES_BY_PATH

        stop_services()

        for service in list(SERVICES_BY_PATH.values()):
            service.join()

        from ubo_app.setup import clear_signal_handlers

        clear_signal_handlers()
        for cleanup in logger_cleanups:
            cleanup()


if __name__ == '__main__':
    main()



# ruff: noqa: D100, D101
from __future__ import annotations

import socket
from dataclasses import field
from datetime import UTC, datetime
from enum import StrEnum, auto
from typing import TYPE_CHECKING, Any
from uuid import uuid4

from immutable import Immutable
from redux import BaseAction, BaseEvent
from ubo_gui.menu.types import ActionItem

from ubo_app.colors import SECONDARY_COLOR_LIGHT
from ubo_app.constants import NOTIFICATIONS_FLASH_TIME
from ubo_app.store.ubo_actions import UboApplicationItem, UboDispatchItem
from ubo_app.utils.dataclass import default_provider

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence

    from kivy.graphics.context_instructions import Color

    from ubo_app.store.services.speech_synthesis import ReadableInformation


class Importance(StrEnum):
    CRITICAL = auto()
    HIGH = auto()
    MEDIUM = auto()
    LOW = auto()


IMPORTANCE_ICONS = {
    Importance.CRITICAL: '󰅚',
    Importance.HIGH: '󰀪',
    Importance.MEDIUM: '',
    Importance.LOW: '󰌶',
}

IMPORTANCE_COLORS = {
    Importance.CRITICAL: '#D32F2F',
    Importance.HIGH: '#FFA000',
    Importance.MEDIUM: '#FFEB3B',
    Importance.LOW: '#2196F3',
}


class NotificationDisplayType(StrEnum):
    NOT_SET = auto()
    BACKGROUND = auto()
    FLASH = auto()
    STICKY = auto()


class Chime(StrEnum):
    ADD = 'add'
    DONE = 'done'
    FAILURE = 'failure'
    VOLUME_CHANGE = 'volume'


class NotificationActionItem(ActionItem):
    background_color: Color | Callable[[], Color] = SECONDARY_COLOR_LIGHT
    dismiss_notification: bool = False
    close_notification: bool = True


class NotificationApplicationItem(UboApplicationItem):
    background_color: Color | Callable[[], Color] = SECONDARY_COLOR_LIGHT
    dismiss_notification: bool = False
    close_notification: bool = True


class NotificationDispatchItem(UboDispatchItem, NotificationActionItem): ...


class Notification(Immutable):
    id: str = field(default_factory=lambda: uuid4().hex)
    title: str
    content: str
    extra_information: ReadableInformation | None = None
    importance: Importance = Importance.LOW
    chime: Chime | None = None
    timestamp: datetime = field(default_factory=lambda: datetime.now(tz=UTC))
    is_read: bool = False
    sender: str | None = None
    actions: Sequence[
        NotificationActionItem | NotificationDispatchItem | NotificationApplicationItem
    ] = field(
        default_factory=list,
    )
    icon: str = field(
        default_factory=default_provider(
            ['importance'],
            lambda importance: IMPORTANCE_ICONS[importance],
        ),
    )
    color: str = field(
        default_factory=default_provider(
            ['importance'],
            lambda importance: IMPORTANCE_COLORS[importance],
        ),
    )
    expiration_timestamp: datetime | None = None
    display_type: NotificationDisplayType = NotificationDisplayType.NOT_SET
    flash_time: float = NOTIFICATIONS_FLASH_TIME
    show_dismiss_action: bool = True
    dismiss_on_close: bool = False
    on_close: Callable[[], Any] | None = None
    blink: bool = True
    progress: float | None = None
    progress_weight: float = 1

    def __post_init__(self) -> None:
        """Replace `{{hostname}}` with the current hostname."""
        object.__setattr__(
            self,
            'title',
            self.title.replace(
                '{{hostname}}',
                f'{socket.gethostname()}.local',
            ),
        )
        object.__setattr__(
            self,
            'content',
            self.content.replace(
                '{{hostname}}',
                f'{socket.gethostname()}.local',
            ),
        )


class NotificationsAction(BaseAction): ...


class NotificationsAddAction(NotificationsAction):
    notification: Notification


class NotificationsDisplayAction(NotificationsAction):
    notification: Notification
    index: int | None = None
    count: int | None = None


class NotificationsClearAction(NotificationsAction):
    notification: Notification


class NotificationsClearByIdAction(NotificationsAction):
    id: str


class NotificationsClearAllAction(NotificationsAction): ...


class NotificationsEvent(BaseEvent): ...


class NotificationsClearEvent(NotificationsEvent):
    notification: Notification


class NotificationsDisplayEvent(NotificationsEvent):
    notification: Notification
    index: int | None = None
    count: int | None = None


class NotificationsState(Immutable):
    notifications: Sequence[Notification]
    unread_count: int
    progress: float | None = None



# ruff: noqa: SLF001, D100, D103
from __future__ import annotations

from datetime import UTC, datetime
from enum import Enum
from typing import TYPE_CHECKING, TypeAlias, TypeVar, cast, overload

import betterproto
import betterproto.casing

import ubo_bindings.ubo.v1

if TYPE_CHECKING:
    from immutable import Immutable

ReturnType: TypeAlias = (
    betterproto.Message
    | betterproto.Enum
    | int
    | float
    | str
    | bytes
    | bool
    | None
    | list['ReturnType']
)


def get_class(object_: Immutable) -> type[betterproto.Message]:
    return getattr(
        ubo_bindings.ubo.v1,
        betterproto.casing.pascal_case(type(object_).__name__),
    )


def get_enum(object_: Enum) -> type[betterproto.Enum]:
    return getattr(
        ubo_bindings.ubo.v1,
        betterproto.casing.pascal_case(type(object_).__name__),
    )


T = TypeVar('T', bound=betterproto.Message)

GRPCSerializable: TypeAlias = 'Enum | Immutable | datetime | None'


@overload
def build_message(
    object_: GRPCSerializable,
    expected_type: type[T],
) -> T: ...
@overload
def build_message(
    object_: GRPCSerializable,
) -> ReturnType: ...
def build_message(  # noqa: C901
    object_: GRPCSerializable,
    expected_type: type[T] | None = None,
) -> ReturnType | T:
    if isinstance(object_, datetime):
        return object_.astimezone(UTC).timestamp()

    if (expected_type and issubclass(expected_type, betterproto.Enum)) or isinstance(
        object_,
        Enum,
    ):
        if not isinstance(object_, Enum):
            msg = f'Expected an Enum, got {type(object_)}'
            raise ValueError(msg)
        if expected_type is None:
            return object_.value
        if not issubclass(expected_type, betterproto.Enum):
            msg = f'Expected a betterproto.Enum, got {expected_type}'
            raise ValueError(msg)
        return getattr(
            expected_type,
            cast('str', 'UNSPECIFIED' if object_ is None else object_.name),
        )

    if isinstance(object_, int | float | str | bytes | bool | None):
        return cast('ReturnType', object_)

    if isinstance(object_, list | tuple):
        if expected_type:
            if hasattr(
                expected_type,
                '_betterproto',
            ) and expected_type._betterproto.sorted_field_names == ('items',):
                fields = {
                    'items': [
                        build_message(
                            item,
                            expected_type=expected_type._betterproto.cls_by_field[
                                'items'
                            ],
                        )
                        for item in object_
                    ],
                }
                return expected_type(**fields)
            return [
                build_message(item, expected_type=expected_type) for item in object_
            ]
        return [build_message(item) for item in object_]

    keys = object_.__dataclass_fields__.keys()

    message_class = get_class(object_)
    if expected_type and (
        message_class is None or not issubclass(message_class, expected_type)
    ):
        msg = f'Expected {expected_type}, got {message_class}'
        raise ValueError(msg)

    fields = {
        betterproto.casing.snake_case(key): build_message(
            getattr(object_, key),
            expected_type=message_class._betterproto.cls_by_field[key],
        )
        for key in keys
    }

    if message_class is None:
        msg = f'Class not found for {object_}'
        raise ValueError(msg)

    if issubclass(message_class, betterproto.Message):
        return message_class(**fields)

    msg = f'Building message from {object_} is not implemented yet'
    raise NotImplementedError(msg)



"""Set up a hotspot on the UBO."""

import subprocess

from ubo_app.logger import get_logger

logger = get_logger('system-manager')


def _start() -> str:
    logger.info('Starting the hotspot')
    try:
        subprocess.run(
            ['/usr/bin/env', 'systemctl', 'start', 'ubo-hotspot'],
            check=True,
        )
    except subprocess.CalledProcessError:
        logger.exception('Failed to start the hotspot properly, stopping it...')
        _stop()
        return 'failed'
    else:
        logger.info('Hotspot started')
        return 'done'


def _stop() -> str:
    logger.info('Stopping the hotspot')
    try:
        subprocess.run(
            ['/usr/bin/env', 'systemctl', 'stop', 'ubo-hotspot'],
            check=False,
        )
    except subprocess.CalledProcessError:
        logger.exception('Failed to stop the hotspot properly')
        return 'failed'
    else:
        logger.info('Hotspot stopped')
        return 'done'


def hotspot_handler(desired_state: str) -> str:
    """Set up a hotspot on the UBO."""
    if desired_state == 'start':
        return _start()
    if desired_state == 'stop':
        return _stop()
    return 'unknown command'



"""Menu definitions for the Ubo GUI."""

from __future__ import annotations

import math
import socket
from datetime import UTC, datetime
from typing import TYPE_CHECKING

from redux import AutorunOptions
from ubo_gui.menu.types import ActionItem, HeadlessMenu, SubMenuItem

from ubo_app.store.core.types import (
    SETTINGS_ICONS,
    PowerOffAction,
    RebootAction,
    SettingsCategory,
)
from ubo_app.store.main import store
from ubo_app.store.services.notifications import (
    Notification,
    NotificationsDisplayAction,
)
from ubo_app.store.settings.menu import SYSTEM_MENU
from ubo_app.store.ubo_actions import UboDispatchItem
from ubo_app.store.update_manager.utils import open_about_menu

if TYPE_CHECKING:
    from collections.abc import Sequence

    from ubo_gui.menu.types import Item


APPS_MENU = HeadlessMenu(
    title='󰀻Docker Apps',
    items=[],
    placeholder='No apps',
)

SETTINGS_MENU = HeadlessMenu(
    title='Settings',
    items=[
        SubMenuItem(
            key=category,
            label=category,
            icon=SETTINGS_ICONS[category],
            sub_menu=HeadlessMenu(
                title=SETTINGS_ICONS[category] + category,
                items=SYSTEM_MENU if category is SettingsCategory.SYSTEM else [],
                placeholder='No settings in this category',
            ),
        )
        for category in SettingsCategory
    ],
)

MAIN_MENU = HeadlessMenu(
    title='󰍜Main',
    items=[
        SubMenuItem(
            key='apps',
            label='Apps',
            icon='󰀻',
            sub_menu=APPS_MENU,
        ),
        SubMenuItem(
            key='settings',
            label='Settings',
            icon='',
            sub_menu=SETTINGS_MENU,
        ),
        ActionItem(
            key='about',
            label='About',
            icon='',
            action=open_about_menu,
        ),
    ],
)


@store.autorun(
    lambda state: state.notifications.unread_count,
    options=AutorunOptions(default_value='Notifications (not loaded)'),
)
def _notifications_title(unread_count: int) -> str:
    return f'Notifications ({unread_count})'


@store.autorun(
    lambda state: state.notifications.notifications,
    options=AutorunOptions(default_value=[]),
)
def notifications_menu_items(notifications: Sequence[Notification]) -> list[Item]:
    """Return a list of menu items for the notification manager."""
    return [
        UboDispatchItem(
            key=str(notification.id),
            label=notification.title,
            icon=notification.icon,
            color='black',
            background_color=notification.color,
            store_action=NotificationsDisplayAction(
                notification=notification,
                index=index,
                count=len(notifications),
            ),
            progress=None
            if notification.progress is None or math.isnan(notification.progress)
            else notification.progress,
        )
        for index, notification in enumerate(notifications)
        if notification.expiration_timestamp is None
        or notification.expiration_timestamp > datetime.now(tz=UTC)
    ]


@store.autorun(
    lambda state: len(state.notifications.notifications),
    options=AutorunOptions(default_value='white'),
)
def _notifications_color(unread_count: int) -> str:
    return 'yellow' if unread_count > 0 else 'white'


HOME_MENU = HeadlessMenu(
    title=f'󰋜{socket.gethostname()}.local',
    items=[
        SubMenuItem(
            key='main',
            label='',
            sub_menu=MAIN_MENU,
            icon='󰍜',
            is_short=True,
        ),
        SubMenuItem(
            key='notifications',
            label='',
            sub_menu=HeadlessMenu(
                title=_notifications_title,
                items=notifications_menu_items,
                placeholder='No notifications',
            ),
            color=_notifications_color,
            icon='',
            is_short=True,
        ),
        SubMenuItem(
            key='power',
            label='',
            sub_menu=HeadlessMenu(
                title='󰐥Power',
                items=[
                    UboDispatchItem(
                        label='Reboot',
                        store_action=RebootAction(),
                        icon='󰜉',
                    ),
                    UboDispatchItem(
                        label='Power off',
                        store_action=PowerOffAction(),
                        icon='󰐥',
                    ),
                ],
            ),
            icon='󰐥',
            is_short=True,
        ),
    ],
)



"""gRPC server for the store service."""

from __future__ import annotations

from grpclib.reflection.service import ServerReflection
from grpclib.server import Server
from ubo_app.constants import GRPC_LISTEN_ADDRESS, GRPC_LISTEN_PORT
from ubo_app.logger import logger
from ubo_app.rpc.secrets_service import SecretsService
from ubo_app.rpc.store_service import StoreService


async def serve() -> None:
    """Serve the gRPC server."""
    services = [StoreService(), SecretsService()]
    services = ServerReflection.extend(services)

    server = Server(services)

    logger.info(
        'Starting gRPC server',
        extra={'host': GRPC_LISTEN_ADDRESS, 'port': GRPC_LISTEN_PORT},
    )
    await server.start(GRPC_LISTEN_ADDRESS, GRPC_LISTEN_PORT)

    await server.wait_closed()



"""The types used in the proto generation."""

from __future__ import annotations

import functools
from typing import Literal, Self, get_args

import betterproto.casing
from immutable import Immutable

FieldType = Literal['string', 'int64', 'float', 'bool', 'bytes', 'None']


global_messages: dict[str, tuple[str, list[tuple[str, _Type]]]] = {}
global_enums: dict[str, str] = {}
global_types: dict[str, str] = {}


class _Type(Immutable):
    """Root class for all types."""

    def get_proto(self, name: str, *, current_package: str | None) -> str:
        """Get the proto representation of the type."""
        _ = name, current_package
        raise NotImplementedError

    def get_embedded_proto(self, name: str) -> str:
        """Get the proto representation of the type when embedded."""
        return self.get_proto(name, current_package=self.package)

    @property
    def as_basic_type(self) -> str | None:
        """Get the basic type of the type."""
        return None

    @property
    def local_name(self) -> str:
        """Get the local name of the type."""
        raise NotImplementedError

    def get_definitions(self, name: str, *, current_package: str | None) -> str:
        """Get the definitions of the type."""
        _ = name, current_package
        raise NotImplementedError

    def get_embedded_definitions(
        self,
        name: str,
        *,
        current_package: str | None,
    ) -> str:
        """Get the definitions of the type when embedded."""
        return self.get_definitions(name, current_package=current_package)

    @property
    def package(self) -> str | None:
        """Get the package of the type."""
        return None


class _BasicType(_Type):
    BASIC_TYPES = ('string', 'int64', 'float', 'bool', 'bytes', 'None')
    type: FieldType | str

    def get_proto(self, name: str | None = None, *, current_package: str | None) -> str:
        _ = name, current_package
        if self.type in ('BaseAction', 'UboAction'):
            return 'Action'
        if self.type == 'UboEvent':
            return 'Event'
        if self.type == 'PageWidget':  # Not supported
            return 'string'
        if self.type == 'Color':  # Assuming it is kivy color
            return 'string'
        """ This version is for the case where we have a separate proto file for each
        package, it is currently not possible because proto files do not support import
        loops, and Action needs NotificationsAddAction which needs
        NotificationActionItem which needs Action

        if self.package is None or self.package == current_package:
            if self.type in global_messages:
                message = global_messages[self.type]
                return f'{message[0]}.v1.{self.type}'
            if self.type in global_enums:
                return f'{global_enums[self.type]}.v1.{self.type}'
            if self.type in global_types:
                return f'{global_types[self.type]}.v1.{self.type}'
            return self.type
        return f'{self.package}.v1.{self.type}'
        """
        return self.type

    @property
    def as_basic_type(self) -> str:
        return self.type

    @property
    def local_name(self) -> str:
        return self.type

    def get_definitions(self, name: str, *, current_package: str | None) -> str:
        _ = name, current_package
        return ''

    @property
    def package(self) -> str | None:
        if self.type not in get_args(FieldType):
            if self.type in global_messages:
                return global_messages[self.type][0]
            if self.type in global_enums:
                return global_enums[self.type]
            if self.type in global_types:
                return global_types[self.type]
            if self.type in ('Color', 'UboAction', 'UboEvent', 'BaseAction'):
                return None
            msg = f'Unknown type "{self.type}"'
            raise TypeError(msg)
        return None


class _OptionalType(_Type):
    type: _Type

    def get_proto(self, name: str, *, current_package: str | None) -> str:
        _ = current_package
        if name == 'prompt':
            pass
        if isinstance(self.type, _BasicType):
            if name == 'prompt':
                pass
            return (
                f'optional {self.type.get_proto(name, current_package=current_package)}'
            )
        return f"""optional {
            self.type.get_embedded_proto(
                betterproto.casing.pascal_case(name),
            )
        }"""

    def get_embedded_proto(self, name: str) -> str:
        return betterproto.casing.pascal_case(name)

    @property
    def local_name(self) -> str:
        return 'optional'

    def get_definitions(self, name: str, *, current_package: str | None) -> str:
        return f"""{
            self.type.get_embedded_definitions(
                f'{betterproto.casing.pascal_case(name)}',
                current_package=current_package,
            )
        }"""

    def get_embedded_definitions(
        self,
        name: str,
        *,
        current_package: str | None,
    ) -> str:
        return f"""{
            self.type.get_embedded_definitions(
                f'{betterproto.casing.pascal_case(name)}Optional',
                current_package=current_package,
            )
        }
message {betterproto.casing.pascal_case(name)} {{
  optional {
            self.type.get_embedded_proto(
                f'{betterproto.casing.pascal_case(name)}Optional',
            )
        } items = 1;
}}
"""


class _ListType(_Type):
    type: _Type

    def get_proto(self, name: str, *, current_package: str | None) -> str:
        _ = current_package
        return f"""repeated {
            self.type.get_embedded_proto(
                betterproto.casing.pascal_case(name),
            )
        }"""

    def get_embedded_proto(self, name: str) -> str:
        return betterproto.casing.pascal_case(name)

    @property
    def local_name(self) -> str:
        return 'list'

    def get_definitions(self, name: str, *, current_package: str | None) -> str:
        return f"""{
            self.type.get_embedded_definitions(
                f'{betterproto.casing.pascal_case(name)}',
                current_package=current_package,
            )
        }"""

    def get_embedded_definitions(
        self,
        name: str,
        *,
        current_package: str | None,
    ) -> str:
        return f"""{
            self.type.get_embedded_definitions(
                f'{betterproto.casing.pascal_case(name)}Item',
                current_package=current_package,
            )
        }
message {betterproto.casing.pascal_case(name)} {{
  repeated {
            self.type.get_embedded_proto(
                f'{betterproto.casing.pascal_case(name)}Item',
            )
        } items = 1;
}}
"""


class _SetType(_Type):
    type: _Type

    def get_proto(self, name: str, *, current_package: str | None) -> str:
        _ = current_package
        return self.get_embedded_proto(name)

    def get_embedded_proto(self, name: str) -> str:
        return betterproto.casing.pascal_case(name) + 'SetType'

    @property
    def local_name(self) -> str:
        return 'set'

    def get_definitions(self, name: str, *, current_package: str | None) -> str:
        return self.get_embedded_definitions(
            name,
            current_package=current_package,
        )

    def get_embedded_definitions(
        self,
        name: str,
        *,
        current_package: str | None,
    ) -> str:
        return f"""{
            self.type.get_embedded_definitions(
                f'{betterproto.casing.pascal_case(name)}Item',
                current_package=current_package,
            )
        }
message {betterproto.casing.pascal_case(name)}SetType {{
  repeated {
            self.type.get_embedded_proto(
                f'{betterproto.casing.pascal_case(name)}Item',
            )
        } items = 1;
}}
"""


class _UnionType(_Type):
    types: tuple[_Type, ...]

    @classmethod
    def from_(cls: type[Self], *, types: list[_Type]) -> _Type:
        types_ = functools.reduce(
            lambda x, y: (*x, *y),
            (item.types if isinstance(item, _UnionType) else (item,) for item in types),
            (),
        )

        if len(types_) == 1:
            return types_[0]

        return cls(types=types_)

    def get_proto(self, name: str, *, current_package: str | None) -> str:
        _ = current_package
        return betterproto.casing.pascal_case(name)

    def get_embedded_proto(self, name: str) -> str:
        return betterproto.casing.pascal_case(name)

    @property
    def local_name(self) -> str:
        return 'union'

    def get_definitions(self, name: str, *, current_package: str | None) -> str:
        return self.get_embedded_definitions(name, current_package=current_package)

    def get_embedded_definitions(
        self,
        name: str,
        *,
        current_package: str | None,
    ) -> str:
        sub_definitions = ''
        definitions = f'message {betterproto.casing.pascal_case(name)} {{\n'
        if len(self.types) > 0:
            definitions += f'  oneof {betterproto.casing.snake_case(name)} {{\n'
            index = 1
            for item in sorted(self.types, key=lambda x: x.local_name):
                try:
                    definitions += f"""  {item.get_embedded_proto(f'{name}_{index}')} {
                        betterproto.casing.snake_case(item.local_name)
                    } = {index};\n"""
                    sub_definitions += item.get_embedded_definitions(
                        f'{name}_{index}',
                        current_package=current_package,
                    )
                except TypeError as exception:
                    if 'Unknown type' in str(exception) or 'Empty Union' in str(
                        exception,
                    ):
                        continue
                else:
                    index += 1
            definitions += '  }\n'
        definitions += '}\n'

        return f'{sub_definitions}\n{definitions}'


class _DictType(_Type):
    key_type: _Type
    value_type: _Type

    def get_proto(self, name: str, *, current_package: str | None) -> str:
        _ = current_package
        return f"""map<{
            'string'
            if isinstance(self.key_type, _BasicType)
            and self.key_type.type in global_enums
            else self.key_type.get_embedded_proto(f'{name}_key')
        }, {self.value_type.get_embedded_proto(f'{name}_value')}>"""

    def get_embedded_proto(self, name: str) -> str:
        return f'{betterproto.casing.pascal_case(name)}Dict'

    @property
    def local_name(self) -> str:
        return 'dict'

    def get_definitions(self, name: str, *, current_package: str | None) -> str:
        return f"""{
            self.key_type.get_definitions(
                f'{name}_key',
                current_package=current_package,
            )
        }\n\n{
            self.value_type.get_definitions(
                f'{name}_value',
                current_package=current_package,
            )
        }"""

    def get_embedded_definitions(
        self,
        name: str,
        *,
        current_package: str | None,
    ) -> str:
        return f"""{
            self.key_type.get_embedded_definitions(
                f'{name}_key',
                current_package=current_package,
            )
        }

{
            self.value_type.get_embedded_definitions(
                f'{name}_value',
                current_package=current_package,
            )
        }
message {betterproto.casing.pascal_case(name)}Dict {{
  {self.get_proto(name, current_package=current_package)} items = 1;
}}"""


_ = [_BasicType, _OptionalType, _ListType, _SetType, _UnionType, _DictType]



# ruff: noqa: D100, D103
from __future__ import annotations

from dataclasses import replace

from redux import (
    BaseEvent,
    CompleteReducerResult,
    InitAction,
    InitializationActionError,
)

from ubo_app.colors import SUCCESS_COLOR
from ubo_app.store.services.notifications import (
    Importance,
    Notification,
    NotificationDisplayType,
    NotificationsAddAction,
)
from ubo_app.store.services.rpi_connect import (
    RPiConnectAction,
    RPiConnectDoneDownloadingAction,
    RPiConnectLoginEvent,
    RPiConnectSetPendingAction,
    RPiConnectSetStatusAction,
    RPiConnectStartDownloadingAction,
    RPiConnectState,
    RPiConnectUpdateServiceStateAction,
)


def reducer(
    state: RPiConnectState | None,
    action: RPiConnectAction,
) -> (
    CompleteReducerResult[RPiConnectState, NotificationsAddAction, BaseEvent]
    | RPiConnectState
):
    if state is None:
        if isinstance(action, InitAction):
            return RPiConnectState()
        raise InitializationActionError(action)

    match action:
        case RPiConnectStartDownloadingAction():
            return replace(state, is_downloading=True)

        case RPiConnectDoneDownloadingAction():
            return replace(state, is_downloading=False)

        case RPiConnectSetPendingAction():
            return replace(state, is_installed=None, is_signed_in=None, status=None)

        case RPiConnectUpdateServiceStateAction():
            if action.is_active is not None:
                state = replace(state, is_active=action.is_active)
            return state

        case RPiConnectSetStatusAction():
            actions = []
            events = []
            if state.is_signed_in is False and action.is_signed_in:
                actions.append(
                    NotificationsAddAction(
                        notification=Notification(
                            title='RPi-Connect',
                            content='Successful Login',
                            icon='',
                            importance=Importance.MEDIUM,
                            color=SUCCESS_COLOR,
                            display_type=NotificationDisplayType.FLASH,
                        ),
                    ),
                )
                events.append(RPiConnectLoginEvent())

            state = replace(
                state,
                is_installed=action.is_installed,
                is_signed_in=action.is_signed_in,
                status=action.status,
            )

            return CompleteReducerResult(state=state, actions=actions, events=events)

        case _:
            return state



"""Service menu items."""

from __future__ import annotations

import datetime
import logging
from typing import TYPE_CHECKING

from ubo_gui.menu.types import (
    HeadedMenu,
    HeadlessMenu,
    Item,
    SubMenuItem,
)

from ubo_app import logger
from ubo_app.colors import (
    DANGER_COLOR,
    RUNNING_COLOR,
    STOPPED_COLOR,
    WARNING_COLOR,
)
from ubo_app.store.main import store
from ubo_app.store.settings.types import (
    ErrorReport,
    SettingsClearServiceErrorsAction,
    SettingsServiceSetIsEnabledAction,
    SettingsServiceSetLogLevelAction,
    SettingsServiceSetShouldRestartAction,
    SettingsStartServiceAction,
    SettingsStopServiceAction,
)
from ubo_app.store.ubo_actions import (
    UboApplicationItem,
    UboDispatchItem,
)
from ubo_app.utils.gui import (
    SELECTED_ITEM_PARAMETERS,
    UNSELECTED_ITEM_PARAMETERS,
    ItemParameters,
)


def _get_selected_item_parameters(log_level_: int) -> ItemParameters:
    return {
        **SELECTED_ITEM_PARAMETERS,
        'background_color': logger.COLORS_HEX[log_level_],
        'color': '#ffffff',
    }


def _get_unselected_item_parameters(log_level_: int) -> ItemParameters:
    return {
        **UNSELECTED_ITEM_PARAMETERS,
        'background_color': '#000000',
        'color': logger.COLORS_HEX[log_level_],
    }


if TYPE_CHECKING:
    from ubo_app.store.settings.types import ServiceState


SERVICE_ITEMS: dict[str, SubMenuItem] = {}


def _create_service_item(service: ServiceState) -> SubMenuItem:  # noqa: C901
    if service.id in SERVICE_ITEMS:
        return SERVICE_ITEMS[service.id]

    @store.autorun(
        lambda state: (
            state.settings.services[service.id].label,
            state.settings.services[service.id].is_active,
            state.settings.services[service.id].errors,
        )
        if state.settings.services is not None and service.id in state.settings.services
        else None,
    )
    def heading(data: tuple[str, bool, list] | None) -> str:
        if data is None:
            return ''
        label, is_active, errors = data
        return (
            (
                f'[color={WARNING_COLOR}]󰪥[/color] {label}'
                if errors
                else f'[color={RUNNING_COLOR}]󰪥[/color] {label}'
            )
            if is_active
            else f'[color={STOPPED_COLOR}]󰝦[/color] {label}'
        )

    @store.autorun(
        lambda state: state.settings.services[service.id].errors
        if state.settings.services is not None and service.id in state.settings.services
        else None,
    )
    def sub_heading(errors: list[ErrorReport] | None) -> str:
        if errors is None:
            return ''

        if not errors:
            return 'No errors raised in this service'

        if len(errors) == 1:
            return '1 error raised in this service'

        return f'{len(errors)} errors raised in this service'

    @store.autorun(
        lambda state: state.settings.services[service.id].errors,
    )
    def error_items(errors: list[ErrorReport]) -> list[Item]:
        return [
            UboApplicationItem(
                key=str(index),
                label=datetime.datetime.fromtimestamp(
                    error.timestamp,
                )
                .astimezone()
                .strftime('%Y-%m-%d %H:%M:%S'),
                icon=f'[color={DANGER_COLOR}][/color]',
                application_id='ubo:raw-text-viewer',
                initialization_kwargs={
                    'text': error.message,
                },
            )
            for index, error in enumerate(errors)
        ]

    @store.autorun(
        lambda state: state.settings.services[service.id].log_level,
    )
    def log_level_items(log_level: int) -> list[Item]:
        def create_log_level_item(level: int) -> Item:
            selection_parameters = (
                _get_selected_item_parameters(level)
                if level == log_level
                else _get_unselected_item_parameters(level)
            )
            return UboDispatchItem(
                key=logging.getLevelName(level),
                label=logging.getLevelName(level),
                store_action=SettingsServiceSetLogLevelAction(
                    service_id=service.id,
                    log_level=level,
                ),
                **selection_parameters,
            )

        return [create_log_level_item(level) for level in logger.COLORS_HEX]

    @store.autorun(
        lambda state: state.settings.services[service.id].log_level,
    )
    def log_level_title(log_level: int) -> str:
        return f'Log Level: {logging.getLevelName(log_level)}'

    @store.autorun(
        lambda state: state.settings.services[service.id]
        if state.settings.services is not None and service.id in state.settings.services
        else None,
    )
    def items(
        service_state: ServiceState | None,
    ) -> list[Item]:
        if service_state is None:
            return []
        items: list[Item] = []

        if service_state.is_active:
            items.append(
                UboDispatchItem(
                    key='stop',
                    label='Stop',
                    store_action=SettingsStopServiceAction(service_id=service.id),
                    icon='',
                    background_color=DANGER_COLOR,
                ),
            )
        else:
            items.append(
                UboDispatchItem(
                    key='start',
                    label='Start',
                    store_action=SettingsStartServiceAction(service_id=service.id),
                    icon=f'[color={RUNNING_COLOR}][/color]',
                ),
            )

        if service_state.is_enabled:
            items.extend(
                [
                    UboDispatchItem(
                        key='enabled',
                        label='Auto Load',
                        store_action=SettingsServiceSetIsEnabledAction(
                            service_id=service.id,
                            is_enabled=False,
                        ),
                        icon='',
                    ),
                    SubMenuItem(
                        key='log_level',
                        label=f'Level: {logging.getLevelName(service_state.log_level)}',
                        background_color=logger.COLORS_HEX[service_state.log_level],
                        icon='',
                        sub_menu=HeadlessMenu(
                            title=log_level_title,
                            items=log_level_items,
                        ),
                    ),
                ],
            )
        else:
            items.append(
                UboDispatchItem(
                    key='enabled',
                    label='Auto Load',
                    store_action=SettingsServiceSetIsEnabledAction(
                        service_id=service.id,
                        is_enabled=True,
                    ),
                    icon='',
                    background_color='#000000',
                ),
            )

        if service_state.should_auto_restart:
            items.append(
                UboDispatchItem(
                    key='auto_restart',
                    label='Auto Restart',
                    store_action=SettingsServiceSetShouldRestartAction(
                        service_id=service.id,
                        should_auto_restart=False,
                    ),
                    icon='󰜉',
                ),
            )
        else:
            items.append(
                UboDispatchItem(
                    key='auto_restart',
                    label='Auto Restart',
                    store_action=SettingsServiceSetShouldRestartAction(
                        service_id=service.id,
                        should_auto_restart=True,
                    ),
                    icon='󰶕',
                    background_color='#000000',
                ),
            )

        if service_state.errors:
            items.extend(
                [
                    SubMenuItem(
                        key='errors',
                        label='Errors',
                        icon=f'[color={DANGER_COLOR}][/color]',
                        sub_menu=HeadedMenu(
                            title='Errors',
                            heading='Errors',
                            sub_heading='Errors raised in this service',
                            items=error_items,
                        ),
                    ),
                    UboDispatchItem(
                        key='clear_errors',
                        label='Clear errors',
                        store_action=SettingsClearServiceErrorsAction(
                            service_id=service.id,
                        ),
                        icon='',
                    ),
                ],
            )

        return items

    @store.autorun(
        lambda state: (
            state.settings.services[service.id].is_active,
            state.settings.services[service.id].errors,
        )
        if state.settings.services is not None and service.id in state.settings.services
        else None,
    )
    def icon(data: tuple[bool, list[ErrorReport]] | None) -> str:
        if data is None:
            return ''
        is_active, errors = data
        return (
            (
                f'[color={WARNING_COLOR}]󰪥[/color]'
                if errors
                else f'[color={RUNNING_COLOR}]󰪥[/color]'
            )
            if is_active
            else f'[color={STOPPED_COLOR}]󰝦[/color]'
        )

    SERVICE_ITEMS[service.id] = SubMenuItem(
        key=service.id,
        label=service.label,
        icon=icon,
        sub_menu=HeadedMenu(
            title=service.label,
            heading=heading,
            sub_heading=sub_heading,
            items=items,
        ),
    )

    return SERVICE_ITEMS[service.id]


@store.autorun(lambda state: state.settings.services)
def service_items(services: dict[str, ServiceState] | None) -> list[SubMenuItem]:
    """Generate the items of the services menu."""
    if services is None:
        return []

    return [
        _create_service_item(service)
        for service in sorted(services.values(), key=lambda x: x.label)
    ]



# ruff: noqa: D100, D103
from __future__ import annotations

from dataclasses import replace

from constants import INTERNET_STATE_ICON_ID, INTERNET_STATE_ICON_PRIORITY
from redux import (
    CompleteReducerResult,
    InitAction,
    InitializationActionError,
    ReducerResult,
)

from ubo_app.colors import DANGER_COLOR
from ubo_app.store.services.ip import (
    IpAction,
    IpSetIsConnectedAction,
    IpState,
    IpUpdateInterfacesAction,
)
from ubo_app.store.status_icons.types import StatusIconsRegisterAction

Action = InitAction | IpAction


def reducer(
    state: IpState | None,
    action: Action,
) -> ReducerResult[IpState, StatusIconsRegisterAction, None]:
    if state is None:
        if isinstance(action, InitAction):
            return IpState(interfaces=[])
        raise InitializationActionError(action)

    match action:
        case IpUpdateInterfacesAction():
            return replace(state, interfaces=action.interfaces)

        case IpSetIsConnectedAction():
            return CompleteReducerResult(
                state=replace(state, is_connected=action.is_connected),
                actions=[
                    StatusIconsRegisterAction(
                        icon='󰖟'
                        if action.is_connected
                        else f'[color={DANGER_COLOR}]󰪎[/color]',
                        priority=INTERNET_STATE_ICON_PRIORITY,
                        id=INTERNET_STATE_ICON_ID,
                    ),
                ],
            )

        case _:
            return state



# ruff: noqa: D100, D101
from __future__ import annotations

from typing import TYPE_CHECKING

from immutable import Immutable
from redux import BaseEvent

if TYPE_CHECKING:
    from ubo_app.store.input.types import WebUIInputDescription


class WebUIEvent(BaseEvent): ...


class WebUIInitializeEvent(WebUIEvent):
    description: WebUIInputDescription


class WebUIStopEvent(WebUIEvent): ...


class WebUIState(Immutable):
    active_inputs: list[WebUIInputDescription]



# ruff: noqa: D100, D103
from __future__ import annotations

from typing import TYPE_CHECKING, Any

from constants import get_signal_icon
from debouncer import DebounceOptions, debounce
from kivy.clock import mainthread
from kivy.properties import StringProperty
from ubo_gui.menu.types import (
    ActionItem,
    HeadlessMenu,
    SubMenuItem,
)
from ubo_gui.prompt import PromptWidget
from wifi_manager import (
    connect_wireless_connection,
    disconnect_wireless_connection,
    forget_wireless_connection,
    get_active_connection_ssid,
    get_active_connection_state,
    get_wifi_device,
)

from ubo_app.store.core.types import CloseApplicationAction
from ubo_app.store.main import store
from ubo_app.store.services.wifi import (
    ConnectionState,
    WiFiConnection,
    WiFiUpdateRequestAction,
)
from ubo_app.store.ubo_actions import UboApplicationItem, register_application
from ubo_app.utils.async_ import create_task
from ubo_app.utils.gui import UboPromptWidget

from .create_wireless_connection import CreateWirelessConnectionPage

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence


class _WiFiConnectionPage(UboPromptWidget):
    ssid: str = StringProperty()
    state: ConnectionState = StringProperty(defaultvalue=ConnectionState.UNKNOWN)

    def first_option_callback(self) -> None:
        if self.state is ConnectionState.CONNECTED:
            create_task(disconnect_wireless_connection())
        elif self.state is ConnectionState.DISCONNECTED:
            create_task(connect_wireless_connection(self.ssid))
        store.dispatch(WiFiUpdateRequestAction(reset=True))

    def second_option_callback(self) -> None:
        create_task(forget_wireless_connection(self.ssid))
        store.dispatch(
            CloseApplicationAction(application_instance_id=self.id),
            WiFiUpdateRequestAction(reset=True),
        )

    def update(self, *_: tuple[Any, ...]) -> None:
        if self.state is ConnectionState.CONNECTED:
            self.first_option_label = 'Disconnect'
            self.first_option_icon = '󰖪'
            self.first_option_color = 'black'
            self.first_option_background_color = (
                PromptWidget.first_option_background_color.defaultvalue
            )
            self.icon = '󰖩'
        elif self.state is ConnectionState.DISCONNECTED:
            self.first_option_label = 'Connect'
            self.first_option_icon = '󰖩'
            self.first_option_color = 'black'
            self.first_option_background_color = (
                PromptWidget.first_option_background_color.defaultvalue
            )
            self.icon = '󰖪'
        elif self.state is ConnectionState.CONNECTING:
            self.first_option_label = 'Connecting...'
            self.first_option_icon = ''
            self.first_option_color = 'white'
            self.first_option_background_color = 'black'
            self.icon = ''
        elif self.state is ConnectionState.UNKNOWN:
            self.first_option_label = ''
            self.first_option_icon = ''
            self.first_option_color = 'white'
            self.first_option_background_color = 'black'
            self.icon = ''

    def __init__(self, **kwargs: object) -> None:
        super().__init__(**kwargs)
        self.title = None
        self.prompt = f'SSID: {self.ssid}'
        self.first_option_is_short = False
        self.second_option_label = 'Delete'
        self.second_option_icon = '󰆴'
        self.second_option_is_short = False

        self.bind(state=self.update)
        self.update()

        @debounce(
            wait=0.5,
            options=DebounceOptions(leading=False, trailing=True, time_window=0.5),
        )
        async def update_status() -> None:
            if await get_active_connection_ssid() == self.ssid:
                state = await get_active_connection_state()
            else:
                state = ConnectionState.DISCONNECTED
            mainthread(lambda: setattr(self, 'state', state))()

        create_task(update_status())

        async def listener() -> None:
            wifi_device = await get_wifi_device()
            if not wifi_device:
                return

            async for _ in wifi_device.properties_changed:
                create_task(update_status())

        create_task(listener())


register_application(
    application=_WiFiConnectionPage,
    application_id='wifi:connection-page',
)
register_application(
    application=CreateWirelessConnectionPage,
    application_id='wifi:create-connection-page',
)


@store.autorun(lambda state: state.wifi.connections)
def wireless_connections_menu(
    connections: Sequence[WiFiConnection] | None,
) -> HeadlessMenu:
    if connections is None:
        return HeadlessMenu(
            title='Wi-Fi',
            items=[],
            placeholder='Loading...',
        )

    icons = {
        ConnectionState.CONNECTED: '󱚽',
        ConnectionState.DISCONNECTED: '󱛅',
        ConnectionState.CONNECTING: '󱛇',
        ConnectionState.UNKNOWN: '󱚵',
    }
    items = (
        [
            UboApplicationItem(
                key=connection.ssid,
                label=connection.ssid,
                application_id='wifi:connection-page',
                icon=get_signal_icon(connection.signal_strength)
                if connection.state == ConnectionState.DISCONNECTED
                else icons[connection.state],
                initialization_kwargs={
                    'ssid': connection.ssid,
                },
            )
            for connection in connections
        ]
        if connections is not None
        else []
    )

    placeholder = 'Loading...' if connections is None else 'No Wi-Fi connections found'

    return HeadlessMenu(
        title='Wi-Fi',
        items=items,
        placeholder=placeholder,
    )


def list_connections() -> Callable[[], HeadlessMenu]:
    store.dispatch(WiFiUpdateRequestAction())
    return wireless_connections_menu


WiFiMainMenu = SubMenuItem(
    label='WiFi',
    icon='󰖩',
    sub_menu=HeadlessMenu(
        title='WiFi Settings',
        items=[
            UboApplicationItem(
                label='Add',
                icon='󱛃',
                application_id='wifi:create-connection-page',
            ),
            ActionItem(
                label='Select',
                icon='󱖫',
                action=list_connections,
            ),
        ],
    ),
)



"""Ubo service utilities."""

from __future__ import annotations

import sys
import threading
import traceback
from typing import TYPE_CHECKING

from ubo_app.utils.thread import UboThread

if TYPE_CHECKING:
    from ubo_app.service_thread import UboServiceThread
    from ubo_app.utils.types import CoroutineRunner


class ServiceUnavailableError(Exception):
    """Raised when the current service can't be determined."""


def get_service() -> UboServiceThread:
    """Get the current service instance."""
    if 'ubo_app.service_thread' in sys.modules:
        from ubo_app.service_thread import SERVICES_BY_PATH, UboServiceThread

        thread = threading.current_thread()

        if isinstance(thread, UboServiceThread):
            return thread

        if isinstance(thread, UboThread) and thread.ubo_service:
            return thread.ubo_service

        stack = traceback.extract_stack()
        services_by_path = SERVICES_BY_PATH.copy()

        # Optimize by checking stack frames in reverse order and breaking early
        for frame in stack[-2::-1]:
            frame_path = frame.filename
            for registered_path in services_by_path:
                if frame_path.startswith(registered_path.as_posix()):
                    if registered_path in SERVICES_BY_PATH:
                        return SERVICES_BY_PATH[registered_path]
                    break  # Move to next frame if path not found in current services

    msg = 'Service is not available.'
    raise ServiceUnavailableError(msg)


def get_coroutine_runner() -> CoroutineRunner:
    """Get the current service's coroutine runner."""
    try:
        return get_service().run_coroutine
    except ServiceUnavailableError:
        from ubo_app.service import run_coroutine

        return run_coroutine



"""Speech recognition mixin abstract base class."""

from __future__ import annotations

import abc
from typing import TYPE_CHECKING, cast, final, overload

from typing_extensions import override

from abstraction.base_class import BaseSpeechRecognitionEngine
from ubo_app.logger import logger
from ubo_app.store.services.speech_recognition import (
    SpeechRecognitionEngineName,
)
from ubo_app.utils.async_evicting_queue import AsyncEvictingQueue

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Sequence


class Recognition:
    """Base class for recognition data."""

    def __init__(self, engine_name: SpeechRecognitionEngineName) -> None:
        """Initialize a recognition instance."""
        self.audio = b''
        self.text = ''
        self.engine_name = engine_name

    @final
    def append_voice(self, data: bytes) -> None:
        """Append a chunk of audio data to the ongoing voice recognition."""
        self.audio += data

    @final
    def append_text(self, text: str) -> None:
        """Append text to the ongoing voice recognition."""
        if self.text:
            self.text += ' '
        self.text += text


class SpeechRecognition(Recognition):
    """Recognition data for ongoing speech recognition."""

    def __init__(
        self,
        name: SpeechRecognitionEngineName,
        *,
        end_phrase: str,
    ) -> None:
        """Initialize a recognition instance."""
        self.end_phrase = end_phrase
        super().__init__(name)


class PhraseRecognition(Recognition):
    """Recognition data for ongoing phrase recognition."""

    def __init__(
        self,
        engine_name: SpeechRecognitionEngineName,
        *,
        phrases: Sequence[str],
    ) -> None:
        """Initialize a recognition instance."""
        self.phrases = phrases
        super().__init__(engine_name)


class SpeechRecognitionMixin(BaseSpeechRecognitionEngine, abc.ABC):
    """Base class for speech recognition engines."""

    @property
    def speech_engine_name(self) -> SpeechRecognitionEngineName:
        """Get the name of the speech recognition engine."""
        return cast('SpeechRecognitionEngineName', self.name)

    @override
    def __init__(self, *, label: str | None = None) -> None:
        """Initialize speech recognition engine."""
        if self.name not in SpeechRecognitionEngineName.__members__.values():
            msg = f'Engine name {self.name} is not a valid SpeechRecognitionEngineName.'
            raise ValueError(msg)
        self.ongoing_recognition: Recognition | None = None
        self.speech_recognitions_queue: AsyncEvictingQueue[Recognition | None] = (
            AsyncEvictingQueue(maxsize=5)
        )
        super().__init__(label=label)

    def should_be_running(self) -> bool:
        """Check if the speech recognition engine should be running."""
        return self.ongoing_recognition is not None or super().should_be_running()

    @overload
    async def activate_speech_recognition(
        self,
        *,
        end_phrase: str,
    ) -> None: ...
    @overload
    async def activate_speech_recognition(
        self,
        *,
        phrases: Sequence[str],
    ) -> None: ...
    @final
    async def activate_speech_recognition(
        self,
        *,
        end_phrase: str | None = None,
        phrases: Sequence[str] | None = None,
    ) -> None:
        """Activate speech recognition."""
        if self.ongoing_recognition is not None:
            msg = 'Speech recognition is already active.'
            raise RuntimeError(msg)
        if end_phrase is not None:
            self.ongoing_recognition = SpeechRecognition(
                self.speech_engine_name,
                end_phrase=end_phrase,
            )
        elif phrases is not None:
            self.ongoing_recognition = PhraseRecognition(
                self.speech_engine_name,
                phrases=phrases,
            )

        self.decide_running_state()

    @final
    async def deactivate_speech_recognition(self) -> None:
        """Deactivate the ongoing speech recognition."""
        self.ongoing_recognition = None
        await self.speech_recognitions_queue.put(None)

        self.decide_running_state()

    @final
    async def _complete_speech_recognition(self) -> None:
        """Complete the ongoing speech recognition."""
        if self.ongoing_recognition is None:
            msg = 'Speech recognition is not active.'
            raise RuntimeError(msg)
        await self.speech_recognitions_queue.put(self.ongoing_recognition)

    @override
    async def report(self, result: str) -> None:
        """Report the recognized speech."""
        logger.info(
            'Speech recognized',
            extra={
                'result': result,
                'engine_name': self.name,
            },
        )
        if self.ongoing_recognition:
            if isinstance(self.ongoing_recognition, SpeechRecognition):
                indices = [
                    result.lower().index(term)
                    for term in self.ongoing_recognition.end_phrase.split()
                    if term in result.lower()
                ]
                if len(indices) == len(
                    self.ongoing_recognition.end_phrase.split(),
                ) and indices == sorted(indices):
                    logger.info(
                        'End phrase detected, completing recognition',
                        extra={
                            'engine_name': self.name,
                            'end_phrase': self.ongoing_recognition.end_phrase,
                        },
                    )
                    self.ongoing_recognition.append_text(result[: indices[0]])
                    await self._complete_speech_recognition()
                else:
                    self.ongoing_recognition.append_text(result)
            elif isinstance(self.ongoing_recognition, PhraseRecognition):
                for phrase in self.ongoing_recognition.phrases:
                    if phrase.lower() in result.lower():
                        logger.info(
                            'Recognized phrase detected, completing recognition',
                            extra={
                                'engine_name': self.name,
                                'phrase': phrase,
                            },
                        )
                        self.ongoing_recognition.append_text(phrase)
                        await self._complete_speech_recognition()
                        break
        else:
            await super().report(result)

    @final
    async def speech_recognitions(self) -> AsyncGenerator[Recognition, None]:
        """Yield recognized speeches."""
        while recognition := await self.speech_recognitions_queue.get():
            yield recognition



# ruff: noqa: D100, D101
from __future__ import annotations

from immutable import Immutable
from redux import BaseAction, BaseEvent


class VSCodeAction(BaseAction): ...


class VSCodeEvent(BaseEvent): ...


class VSCodeStartDownloadingAction(VSCodeAction): ...


class VSCodeDoneDownloadingAction(VSCodeAction): ...


class VSCodeSetPendingAction(VSCodeAction): ...


class VSCodeStatus(Immutable):
    is_service_installed: bool
    is_running: bool
    name: str | None


class VSCodeSetStatusAction(VSCodeAction):
    is_binary_installed: bool
    is_logged_in: bool
    status: VSCodeStatus | None


class VSCodeLoginEvent(VSCodeEvent): ...


class VSCodeRestartEvent(VSCodeEvent): ...


class VSCodeState(Immutable):
    is_pending: bool = True
    is_downloading: bool = False
    is_binary_installed: bool = False
    is_logged_in: bool | None = None
    status: VSCodeStatus | None = None



# ruff: noqa: D100, D103
from __future__ import annotations

from collections.abc import Sequence
from dataclasses import replace
from typing import TYPE_CHECKING

from redux import CompleteReducerResult, InitAction, InitializationActionError

from ubo_app.constants.assistant import ASSISTANT_WAKE_WORD, INTENTS_WAKE_WORD
from ubo_app.store.core.types import (
    MenuChooseByIndexAction,
    MenuGoBackAction,
    MenuGoHomeAction,
    MenuScrollAction,
    MenuScrollDirection,
    OpenApplicationAction,
)
from ubo_app.store.input.types import InputMethod
from ubo_app.store.services.assistant import (
    AssistantStartListeningAction,
)
from ubo_app.store.services.audio import (
    AudioChangeVolumeAction,
    AudioDevice,
    AudioSetMuteStatusAction,
)
from ubo_app.store.services.infrared import InfraredSendCodeAction
from ubo_app.store.services.rgb_ring import (
    RgbRingBlankAction,
    RgbRingCommandAction,
    RgbRingRainbowAction,
    RgbRingSequenceAction,
    RgbRingSetAllAction,
)
from ubo_app.store.services.speech_recognition import (
    SpeechRecognitionAction,
    SpeechRecognitionIntent,
    SpeechRecognitionReportIntentDetectionAction,
    SpeechRecognitionReportSpeechAction,
    SpeechRecognitionReportWakeWordDetectionAction,
    SpeechRecognitionSetIsAssistantActiveAction,
    SpeechRecognitionSetIsIntentsActiveAction,
    SpeechRecognitionSetSelectedEngineAction,
    SpeechRecognitionState,
    SpeechRecognitionStatus,
)

if TYPE_CHECKING:
    from redux import ReducerResult

    from ubo_app.store.main import UboAction


ACKNOWLEDGMENT_ACTION = RgbRingBlankAction()


def reducer(
    state: SpeechRecognitionState | None,
    action: SpeechRecognitionAction,
) -> ReducerResult[SpeechRecognitionState, UboAction, None]:
    if state is None:
        if isinstance(action, InitAction):
            return SpeechRecognitionState(
                intents=[
                    SpeechRecognitionIntent(
                        phrase='Turn on Assistant',
                        action=SpeechRecognitionSetIsAssistantActiveAction(
                            is_active=True,
                        ),
                    ),
                    SpeechRecognitionIntent(
                        phrase='Turn off Assistant',
                        action=SpeechRecognitionSetIsAssistantActiveAction(
                            is_active=False,
                        ),
                    ),
                    SpeechRecognitionIntent(
                        phrase=[
                            'Create WiFi Connection with Camera',
                            'Create WiFi Connection with QR Code',
                            'Create WiFi Connection using Camera',
                            'Create WiFi Connection using QR Code',
                        ],
                        action=OpenApplicationAction(
                            application_id='wifi:create-connection-page',
                            initialization_kwargs={
                                'input_methods': (InputMethod.CAMERA,),
                            },
                        ),
                    ),
                    SpeechRecognitionIntent(
                        phrase=[
                            'Create WiFi Connection with Web Dashboard',
                            'Create WiFi Connection with Web',
                            'Create WiFi Connection with Web UI',
                            'Create WiFi Connection using Web Dashboard',
                            'Create WiFi Connection using Web UI',
                            'Create WiFi Connection using Web',
                        ],
                        action=OpenApplicationAction(
                            application_id='wifi:create-connection-page',
                            initialization_kwargs={
                                'input_methods': (InputMethod.WEB_DASHBOARD,),
                            },
                        ),
                    ),
                    SpeechRecognitionIntent(
                        phrase=['Turn on light strip', 'Turn off light strip'],
                        action=InfraredSendCodeAction(
                            protocol='nec',
                            scancode='0x40',
                        ),
                    ),
                    SpeechRecognitionIntent(
                        phrase='Turn on Lights',
                        action=RgbRingSetAllAction(color=(255, 255, 255)),
                    ),
                    SpeechRecognitionIntent(
                        phrase='Turn off Lights',
                        action=RgbRingSetAllAction(color=(0, 0, 0)),
                    ),
                    SpeechRecognitionIntent(
                        phrase='Turn Lights Red',
                        action=RgbRingSetAllAction(color=(255, 0, 0)),
                    ),
                    SpeechRecognitionIntent(
                        phrase='Turn Lights Green',
                        action=RgbRingSetAllAction(color=(0, 255, 0)),
                    ),
                    SpeechRecognitionIntent(
                        phrase='Turn Lights Blue',
                        action=RgbRingSetAllAction(color=(0, 0, 255)),
                    ),
                    SpeechRecognitionIntent(
                        phrase='Turn Lights Yellow',
                        action=RgbRingSetAllAction(color=(255, 255, 0)),
                    ),
                    SpeechRecognitionIntent(
                        phrase='Turn Lights Purple',
                        action=RgbRingSetAllAction(color=(255, 0, 255)),
                    ),
                    SpeechRecognitionIntent(
                        phrase='Turn Lights Cyan',
                        action=RgbRingSetAllAction(color=(0, 255, 255)),
                    ),
                    SpeechRecognitionIntent(
                        phrase='Turn Lights Orange',
                        action=RgbRingSetAllAction(color=(255, 100, 0)),
                    ),
                    SpeechRecognitionIntent(
                        phrase='Turn Lights White',
                        action=RgbRingSetAllAction(color=(255, 255, 255)),
                    ),
                    SpeechRecognitionIntent(
                        phrase='Turn Lights Rainbow',
                        action=RgbRingRainbowAction(rounds=0, wait=2500),
                    ),
                    SpeechRecognitionIntent(
                        phrase='Turn Volume Up',
                        action=AudioChangeVolumeAction(
                            amount=0.1,
                            device=AudioDevice.OUTPUT,
                        ),
                    ),
                    SpeechRecognitionIntent(
                        phrase='Turn Volume Down',
                        action=AudioChangeVolumeAction(
                            amount=-0.1,
                            device=AudioDevice.OUTPUT,
                        ),
                    ),
                    SpeechRecognitionIntent(
                        phrase='Mute Volume',
                        action=AudioSetMuteStatusAction(
                            is_mute=True,
                            device=AudioDevice.OUTPUT,
                        ),
                    ),
                    SpeechRecognitionIntent(
                        phrase='Unmute Volume',
                        action=AudioSetMuteStatusAction(
                            is_mute=False,
                            device=AudioDevice.OUTPUT,
                        ),
                    ),
                    SpeechRecognitionIntent(
                        phrase='Activate Button One',
                        action=MenuChooseByIndexAction(index=0),
                    ),
                    SpeechRecognitionIntent(
                        phrase='Activate Button Two',
                        action=MenuChooseByIndexAction(index=1),
                    ),
                    SpeechRecognitionIntent(
                        phrase='Activate Button Three',
                        action=MenuChooseByIndexAction(index=1),
                    ),
                    SpeechRecognitionIntent(
                        phrase=['Activate Back Button', 'Go Back'],
                        action=MenuGoBackAction(),
                    ),
                    SpeechRecognitionIntent(
                        phrase=['Activate Home Button', 'Go Home'],
                        action=MenuGoHomeAction(),
                    ),
                    SpeechRecognitionIntent(
                        phrase=['Activate Up Button', 'Scroll Up'],
                        action=MenuScrollAction(direction=MenuScrollDirection.UP),
                    ),
                    SpeechRecognitionIntent(
                        phrase=['Activate Down Button', 'Scroll Down'],
                        action=MenuScrollAction(direction=MenuScrollDirection.DOWN),
                    ),
                ],
            )

        raise InitializationActionError(action)

    match action:
        case SpeechRecognitionSetSelectedEngineAction():
            return replace(
                state,
                selected_engine=action.engine_name,
                status=SpeechRecognitionStatus.IDLE,
            )

        case SpeechRecognitionSetIsIntentsActiveAction():
            return replace(
                state,
                is_intents_active=action.is_active,
                status=SpeechRecognitionStatus.IDLE
                if state.status is SpeechRecognitionStatus.INTENTS_WAITING
                else state.status,
            )

        case SpeechRecognitionSetIsAssistantActiveAction():
            return replace(
                state,
                is_assistant_active=action.is_active,
                status=SpeechRecognitionStatus.IDLE
                if state.status is SpeechRecognitionStatus.ASSISTANT_WAITING
                else state.status,
            )

        case SpeechRecognitionReportWakeWordDetectionAction(
            wake_word=wake_word,
        ):
            if (
                wake_word == INTENTS_WAKE_WORD
                and state.status is SpeechRecognitionStatus.IDLE
            ):
                new_status = (
                SpeechRecognitionStatus.INTENTS_WAITING
            )
                return CompleteReducerResult(
                    state=replace(state, status=new_status),
                    actions=[RgbRingSetAllAction(color=(0, 0, 255))],
                )
            if (
                wake_word == ASSISTANT_WAKE_WORD
                and state.status is SpeechRecognitionStatus.IDLE
            ):
                return CompleteReducerResult(
                    state=replace(state, status=SpeechRecognitionStatus.IDLE),
                    actions=[AssistantStartListeningAction()],
                )
            return CompleteReducerResult(
                state=replace(state, status=SpeechRecognitionStatus.IDLE),
                actions=[],
            )

        case SpeechRecognitionReportIntentDetectionAction():
            actions = (
                action.intent.action
                if isinstance(action.intent.action, Sequence)
                else [action.intent.action]
            )
            rgb_ring_actions = [
                action for action in actions if isinstance(action, RgbRingCommandAction)
            ]
            non_rgb_ring_actions = [
                action
                for action in actions
                if not isinstance(action, RgbRingCommandAction)
            ]
            return CompleteReducerResult(
                state=replace(state, status=SpeechRecognitionStatus.IDLE),
                actions=[
                    RgbRingSequenceAction(
                        sequence=[ACKNOWLEDGMENT_ACTION, *rgb_ring_actions],
                    ),
                    *non_rgb_ring_actions,
                ],
            )

        case SpeechRecognitionReportSpeechAction():
            return CompleteReducerResult(
                state=replace(state, status=SpeechRecognitionStatus.IDLE),
                actions=[ACKNOWLEDGMENT_ACTION],
            )

        case _:
            return state
